import copy
from datetime import datetime
from typing import List
from loguru import logger

from optrabot.optionhelper import OptionHelper
from optrabot.broker.optionpricedata import OptionStrikeData
from optrabot.broker.order import Leg, OptionRight, Order, OrderAction, OrderType, PriceEffect

class DeltaAdjuster:
	from_time: datetime
	credit_legs: List[Leg]
	wing_size: int
	adjustment_orders: List[Order]

	def __init__(self, from_time: datetime, threshold: int) -> None:
		self.from_time = from_time
		self.threshold = threshold
		self.credit_legs = []
		self.wing_size = None
		self.adjustment_orders = []


	def execute(self, trade) -> List[Order] | None:
		"""
		Execute the delta adjustment for the given trade and returns a list of orders to be executed
		to perform the adjustment on the trade.
		"""
		from optrabot.managedtrade import ManagedTrade
		from optrabot.broker.brokerfactory import BrokerFactory
		managed_trade: ManagedTrade = trade
		broker_connector = BrokerFactory().getBrokerConnectorByAccount(trade.template.account)
		assert broker_connector != None

		atm_strike = broker_connector.get_atm_strike(managed_trade.trade.symbol)
		if not atm_strike:
			return
		
		starting_legs: List[Leg] = []
		# Start with current Long legs
		closing_costs = self.calculate_position_price(managed_trade.current_legs, False)
		for leg in managed_trade.current_legs:
			if leg.action == OrderAction.BUY:
				starting_legs.append(copy.copy(leg))
		logger.debug(f'Closing costs for Legs: {closing_costs:.2f}')

		current_short_call_strike = atm_strike
		current_short_put_strike = atm_strike
		option_strike_data = broker_connector.get_option_strike_data(managed_trade.trade.symbol, managed_trade.current_legs[0].expiration)
		strikes = option_strike_data.strikeData.keys()
		
		next_short_call_strike = OptionHelper.get_next_strike(current_short_call_strike, strikes, True)
		next_short_put_strike = OptionHelper.get_next_strike(current_short_put_strike, strikes, False)
		logger.debug(f'ATM Strike: {atm_strike} - starting with Put Strike: {next_short_put_strike} Call Strike: {next_short_call_strike}')
		for leg in managed_trade.current_legs:
			if leg.action == OrderAction.SELL:
				if leg.right == OptionRight.CALL:
					short_call_leg = Leg(action=leg.action, symbol=leg.symbol, quantity=leg.quantity, strike=next_short_call_strike, right=leg.right, expiration=leg.expiration)
					starting_legs.append(short_call_leg)
				elif leg.right == OptionRight.PUT:
					short_put_leg = Leg(action=leg.action, symbol=leg.symbol, quantity=leg.quantity, strike=next_short_put_strike, right=leg.right, expiration=leg.expiration)
					starting_legs.append(short_put_leg)

		# Jetzt den eigentlichen Delta Adjustment Prozess starten.
		self.wing_size = managed_trade.template.get_wing()
		delta_adjusted_legs = self.adjust_delta(starting_legs, strikes, option_strike_data, closing_costs)
		if delta_adjusted_legs == None:
			logger.info(f'Delta adjustment not possible for a credit currently!')
			return

		# Ajdust Long legs if necessary for machting the wing size
		if self.wing_size is not None:
			# Call side
			short_call_leg = next((leg for leg in delta_adjusted_legs if leg.right == OptionRight.CALL and leg.action == OrderAction.SELL), None)
			long_call_leg = next((leg for leg in delta_adjusted_legs if leg.right == OptionRight.CALL and leg.action == OrderAction.BUY), None)
			if long_call_leg.strike - short_call_leg.strike > self.wing_size:
				long_call_leg.strike = short_call_leg.strike + self.wing_size

			# Put side
			short_put_leg = next((leg for leg in delta_adjusted_legs if leg.right == OptionRight.PUT and leg.action == OrderAction.SELL), None)
			long_put_leg = next((leg for leg in delta_adjusted_legs if leg.right == OptionRight.PUT and leg.action == OrderAction.BUY), None)
			if short_put_leg.strike - long_put_leg.strike > self.wing_size:
				long_put_leg.strike = short_put_leg.strike - self.wing_size

		# Compose order operations for performing the adjustment
		strikes_string = ''
		for leg in delta_adjusted_legs:
			if strikes_string != '':
				strikes_string += '/'
			strikes_string += str(leg.strike)
		logger.debug(f'Adjusted Legs to Strikes: {strikes_string}')

		# Detect changes
		original_legs = copy.copy(managed_trade.current_legs)
		original_legs = sorted(original_legs, key=lambda x: (x.action.value, x.right.value))
		long_legs_for_adjustment_order: List[Leg] = []
		short_legs_for_adjustment_order: List[Leg] = []
		for original_leg in original_legs:
			if original_leg.action == OrderAction.BUY:
				# Detect changes in Long Legs
				delta_leg = next((leg for leg in delta_adjusted_legs if leg.right == original_leg.right and leg.action == OrderAction.BUY), None)
				if delta_leg and delta_leg.strike != original_leg.strike:
					logger.info(f'Long Leg Adjusted: {original_leg.strike} -> {delta_leg.strike}')
					if original_leg.bidPrice is not None and original_leg.bidPrice > 0:
						leg_for_closing = copy.copy(original_leg)
						leg_for_closing.action = OrderAction.SELL
						long_legs_for_adjustment_order.append(leg_for_closing)	# First close the original leg
					else:
						logger.debug(f'Long Leg at strike {original_leg.strike} has no bid price...it will left orphan.')
						continue
					# Obtain price data for opening legs
					strike_price_data =  option_strike_data.strikeData[delta_leg.strike]
					delta_leg.askPrice = strike_price_data.putAsk if delta_leg.right == OptionRight.PUT else strike_price_data.callAsk
					delta_leg.bidPrice = strike_price_data.putBid if delta_leg.right == OptionRight.PUT else strike_price_data.callBid
					long_legs_for_adjustment_order.append(copy.copy(delta_leg))
			elif original_leg.action == OrderAction.SELL:
				# Detect changes in Short Legs
				delta_leg = next((leg for leg in delta_adjusted_legs if leg.right == original_leg.right and leg.action == OrderAction.SELL), None)
				if delta_leg and delta_leg.strike != original_leg.strike:
					logger.info(f'Short Leg Adjusted: {original_leg.strike} -> {delta_leg.strike}')
					leg_for_closing = copy.copy(original_leg)
					leg_for_closing.action = OrderAction.BUY
					short_legs_for_adjustment_order.append(leg_for_closing)

					# Obtain price data for opening legs
					strike_price_data =  option_strike_data.strikeData[delta_leg.strike]
					delta_leg.askPrice = strike_price_data.putAsk if delta_leg.right == OptionRight.PUT else strike_price_data.callAsk
					delta_leg.bidPrice = strike_price_data.putBid if delta_leg.right == OptionRight.PUT else strike_price_data.callBid
					
					short_legs_for_adjustment_order.append(copy.copy(delta_leg))
		if len(long_legs_for_adjustment_order) > 0:
			long_leg_adjustment_order = Order(symbol=managed_trade.trade.symbol, legs=long_legs_for_adjustment_order, action=OrderAction.BUY_TO_OPEN, quantity=managed_trade.template.amount, type=OrderType.LIMIT)
			self.adjustment_orders.append(long_leg_adjustment_order)
		if len(short_legs_for_adjustment_order) > 0:
			short_leg_adjustment_order = Order(symbol=managed_trade.trade.symbol, legs=short_legs_for_adjustment_order, action=OrderAction.BUY_TO_OPEN, quantity=managed_trade.template.amount, type=OrderType.LIMIT)
			self.adjustment_orders.append(short_leg_adjustment_order)

		if len(self.adjustment_orders) > 0:
			return self.adjustment_orders
		else:
			return None

	def adjust_delta(self, legs: List[Leg], strikes: list[float], option_strike_data, closing_costs: float) -> List[Leg]:
		"""
		Performs the adjustment of the positions delta recursively.
		"""
		position_delta = self.calculate_position_delta_and_price(legs, option_strike_data) * 100
		if position_delta < 2 and position_delta > -2:
			# Delta is OK, check if price is sufficient
			new_position_price = self.calculate_position_price(legs, True)
			adjustment_cost = closing_costs + new_position_price
			logger.debug(f'Adjusted position opening price: {new_position_price:.2f}')
			logger.debug(f'Adjustment price: {adjustment_cost:.2f}')
			if adjustment_cost >= 0:
				# Store legs which gave a credit
				self.credit_legs = copy.deepcopy(legs)
				# Try to adjust
				return self.adjust_legs(position_delta, legs, strikes, option_strike_data, closing_costs)
			else:
				# Adjustment for a debit --> So we adjusted to much
				return self.credit_legs
		else:
			# Legs need to be adjusted
			return self.adjust_legs(position_delta, legs, strikes, option_strike_data, closing_costs)

	def adjust_legs(self, position_delta: float, legs: List[Leg], strikes: list[float], option_strike_data, closing_costs: float) -> List[Leg]:
		if position_delta > 0:
			# Short Put leg need to be further OTM
			for leg in legs:
				if leg.right == OptionRight.PUT and leg.action == OrderAction.SELL:
					# Find next OTM Put Strike
					next_strike = OptionHelper.get_next_strike(leg.strike, strikes, False)
					if next_strike is not None:
						leg.strike = next_strike
						logger.debug(f'Adjusted Put Leg to Strike: {next_strike}')
					else:
						logger.error('No further OTM Put Strike found, cannot adjust position delta.')
						return legs
		else:
			# Short Call leg needs to be further OTM
			for leg in legs:
				if leg.right == OptionRight.CALL and leg.action == OrderAction.SELL:
					# Find next OTM Call Strike
					next_strike = OptionHelper.get_next_strike(leg.strike, strikes, True)
					if next_strike is not None:
						leg.strike = next_strike
						logger.debug(f'Adjusted Call Leg to Strike: {next_strike}')
					else:
						logger.error('No further OTM Call Strike found, cannot adjust position delta.')
						return legs
		return self.adjust_delta(legs, strikes, option_strike_data, closing_costs)

	def calculate_position_delta_and_price(self, legs: List[Leg], option_strike_data: OptionStrikeData) -> float:
		"""
		Calculates the total delta of the given legs and the mid price of the legs
		"""
		leg_delta = 0.0
		position_delta = 0.0
		for leg in legs:
			if leg.right == OptionRight.CALL:
				leg_delta = option_strike_data.strikeData[leg.strike].get_call_delta() * leg.quantity
				leg.midPrice = option_strike_data.strikeData[leg.strike].getCallMidPrice()
			elif leg.right == OptionRight.PUT:
				leg_delta = option_strike_data.strikeData[leg.strike].get_put_delta() * leg.quantity
				leg.midPrice = option_strike_data.strikeData[leg.strike].getPutMidPrice()
			if leg.action == OrderAction.SELL:
				leg_delta *= -1
			position_delta += leg_delta
		return position_delta

	def calculate_position_price(self, legs: List[Leg], open: bool = True) -> float:
		"""
		Calculates the total price of the given legs.
		- Positive position prices indicate a net credit (income)
		- Negative position prices indicate a net debit (cost)
		"""
		total_price = 0.0
		for leg in legs:
			if leg.action == OrderAction.BUY:
				if open:
					total_price -= leg.midPrice * leg.quantity
				else:
					total_price += leg.midPrice * leg.quantity
			elif leg.action == OrderAction.SELL:
				if open:
					total_price += leg.midPrice * leg.quantity
				else:
					total_price -= leg.midPrice * leg.quantity
		return total_price



from typing import OrderedDict
from datetime import datetime
from loguru import logger
class EarlyExitTriggerType:
	Breakeven: str = 'breakeven'
	Time: str = 'time'

class EarlyExitTrigger:
	_trigger_time: datetime = None
	type: EarlyExitTriggerType = None

	def __init__(self, triggerdata: OrderedDict) -> None:
		type = triggerdata['type']
		assert type in [EarlyExitTriggerType.Breakeven, EarlyExitTriggerType.Time]
		self.type = type
		try:
			self.value = triggerdata['value']
		except KeyError:
			self.value = None
			pass

	def toDict(self):
		""" Returns a dictionary representation of the EarlyExit Trigger which is used for
		the config file.
		"""
		returnDict = {'type': self.type}
		# eastern = pytz.timezone('US/Eastern')
		# if self.fromTimeUTC != None:
		# 	returnDict['timefrom'] = self.fromTimeUTC.astimezone(eastern).strftime('%H:%M %Z')
		# if self.toTimeUTC != None:
		# 	returnDict['timeto'] = self.toTimeUTC.astimezone(eastern).strftime('%H:%M %Z')
		# if self.excludeFromTimeUTC != None:
		# 	returnDict['excludetimefrom'] = self.excludeFromTimeUTC.astimezone(eastern).strftime('%H:%M %Z')
		# if self.excludeToTimeUTC != None:
		# 	returnDict['excludetimeto'] = self.excludeToTimeUTC.astimezone(eastern).strftime('%H:%M %Z')	
		return returnDict


import copy
from datetime import datetime
from typing import List

from loguru import logger

from optrabot.broker.optionpricedata import OptionStrikeData
from optrabot.broker.order import (Leg, OptionRight, Order, OrderAction,
                                   OrderType, PriceEffect)
from optrabot.managedtrade import ManagedTrade
from optrabot.optionhelper import OptionHelper
from optrabot.signaldata import SignalData
from optrabot.tradetemplate.processor.templateprocessorbase import \
    TemplateProcessorBase
from optrabot.tradetemplate.templatefactory import IronCondor, Template


class IronCondorProcessor(TemplateProcessorBase):
	def __init__(self, template: Template):
		"""
		Initializes the Iron Condor Processor with the given template
		"""
		super().__init__(template)
		self.credit_legs = None

	def composeEntryOrder(self, signalData: SignalData = None):
		"""
		Composes the entry order for the Iron Condor template
		"""
		super().composeEntryOrder(signalData)
		iron_condor_template :IronCondor = self._template
		short_strike_put = None
		short_strike_call = None
		premium = iron_condor_template.get_premium()

		# Short Strike Determination
		shortStrikeData = iron_condor_template.getShortStrikeData()
		if not shortStrikeData and not premium:
			raise ValueError('Configuration for Short Strike or premium is missing in template!')

		if iron_condor_template.wing == None:
				raise ValueError('Wing size has not been defined!')

		if shortStrikeData:
			if shortStrikeData.offset:
				logger.debug(f'Using Short Strike Offset: {shortStrikeData.offset}')
				short_strike_put =  OptionHelper.roundToStrikePrice(signalData.close - shortStrikeData.offset)
				short_strike_call =  OptionHelper.roundToStrikePrice(signalData.close + shortStrikeData.offset)
			elif shortStrikeData.delta:
				logger.debug(f'Using Short Strike Delta: {shortStrikeData.delta}')
				try:
					short_strike_put = self.get_short_strike_from_delta("SPX", OptionRight.PUT, shortStrikeData.delta)
					short_strike_call = self.get_short_strike_from_delta("SPX", OptionRight.CALL, shortStrikeData.delta)
				except ValueError as value_error:
					logger.error(f'Error while determining Short Strike by Delta: {value_error}')
					return None

			logger.debug(f'Using Wing Size: {iron_condor_template.wing}')
			long_strike_put = self.get_valid_strike(short_strike_put - iron_condor_template.wing, expiration_date=self._determined_expiration_date, higher=True)
			long_strike_call = self.get_valid_strike(short_strike_call + iron_condor_template.wing, expiration_date=self._determined_expiration_date, higher=False)

			# Now create the entry order with its legs as all required strikes are determined
			legs: List[Leg] = []
			legs.append(Leg(action=OrderAction.SELL, strike=short_strike_put, symbol=self._template.symbol, right=OptionRight.PUT, expiration=self._determined_expiration_date, quantity=1))
			legs.append(Leg(action=OrderAction.SELL, strike=short_strike_call, symbol=self._template.symbol, right=OptionRight.CALL, expiration=self._determined_expiration_date, quantity=1))
			legs.append(Leg(action=OrderAction.BUY, strike=long_strike_call, symbol=self._template.symbol, right=OptionRight.CALL, expiration=self._determined_expiration_date, quantity=1))
			legs.append(Leg(action=OrderAction.BUY, strike=long_strike_put, symbol=self._template.symbol, right=OptionRight.PUT, expiration=self._determined_expiration_date, quantity=1))
		elif premium:
			logger.debug(f'Trying to create an Iron Condor with premium: ${premium} and Wing Size: {iron_condor_template.wing}')
			self.credit_legs = None
			legs = self.compose_legs_with_min_premium(iron_condor_template.symbol, self._determined_expiration_date, premium, iron_condor_template.wing)
			if not legs:
				raise ValueError('Unable to determine legs for Iron Condor based on premium and wing size!')

		entryOrder = Order(symbol=self._template.symbol, legs=legs, action=OrderAction.BUY_TO_OPEN, quantity=self._template.amount, type=OrderType.LIMIT)
		return entryOrder
	
	def calculate_position_delta_and_price(self, legs: List[Leg], option_strike_data: OptionStrikeData) -> float:
		"""
		Calculates the total delta of the given legs and the mid price of the legs
		"""
		leg_delta = 0.0
		position_delta = 0.0
		for leg in legs:
			if leg.right == OptionRight.CALL:
				leg_delta = option_strike_data.strikeData[leg.strike].get_call_delta() * leg.quantity
				leg.midPrice = option_strike_data.strikeData[leg.strike].getCallMidPrice()
			elif leg.right == OptionRight.PUT:
				leg_delta = option_strike_data.strikeData[leg.strike].get_put_delta() * leg.quantity
				leg.midPrice = option_strike_data.strikeData[leg.strike].getPutMidPrice()
			if leg.action == OrderAction.SELL:
				leg_delta *= -1
			position_delta += leg_delta
		return position_delta
	
	def calculate_position_price(self, legs: List[Leg], open: bool = True) -> float:
		"""
		Calculates the total price of the given legs.
		- Positive position prices indicate a net credit (income)
		- Negative position prices indicate a net debit (cost)
		"""
		total_price = 0.0
		for leg in legs:
			if leg.action == OrderAction.BUY:
				if open:
					total_price -= leg.midPrice * leg.quantity
				else:
					total_price += leg.midPrice * leg.quantity
			elif leg.action == OrderAction.SELL:
				if open:
					total_price += leg.midPrice * leg.quantity
				else:
					total_price -= leg.midPrice * leg.quantity
		return total_price
	
	def compose_legs_with_min_premium(self, symbol: str, expiration_date: datetime.date, min_premium: float, wing: float) -> List[Leg] | None:
		"""
		Composes the Iron Condor legs based on the given minimum premium and the wing size.
		"""
		atm_strike = self.broker_connector.get_atm_strike(self._template.symbol)
		if not atm_strike:
			return
		
		current_short_call_strike = atm_strike
		current_short_put_strike = atm_strike
		option_strike_data = self.broker_connector.get_option_strike_data(symbol, expiration_date)
		strikes = option_strike_data.strikeData.keys()
		next_short_call_strike = OptionHelper.get_next_strike(current_short_call_strike, strikes, True)
		next_short_put_strike = OptionHelper.get_next_strike(current_short_put_strike, strikes, False)
		logger.debug(f'ATM Strike: {atm_strike} - starting with Put Strike: {next_short_put_strike} Call Strike: {next_short_call_strike}')

		starting_legs: List[Leg] = []
		starting_legs.append(Leg(action=OrderAction.SELL, strike=next_short_put_strike, symbol=symbol, right=OptionRight.PUT, expiration=expiration_date, quantity=1))
		starting_legs.append(Leg(action=OrderAction.SELL, strike=next_short_call_strike, symbol=symbol, right=OptionRight.CALL, expiration=expiration_date, quantity=1))
		long_strike_put = self.get_valid_strike(next_short_put_strike - self._template.wing, expiration_date, True)
		long_strike_call = self.get_valid_strike(next_short_call_strike + self._template.wing, expiration_date, False)
		starting_legs.append(Leg(action=OrderAction.BUY, strike=long_strike_call, symbol=self._template.symbol, right=OptionRight.CALL, expiration=expiration_date, quantity=1))
		starting_legs.append(Leg(action=OrderAction.BUY, strike=long_strike_put, symbol=self._template.symbol, right=OptionRight.PUT, expiration=expiration_date, quantity=1))

		entry_order_legs = self.determine_order_legs(starting_legs, strikes, option_strike_data, min_premium)

		return entry_order_legs
	
	def determine_order_legs(self, legs: List[Leg], strikes: List[float], option_strike_data, min_premium: float) -> List[Leg] | None:	
		"""
		Determines the according order legs recursively.
		"""
		position_delta = round(self.calculate_position_delta_and_price(legs, option_strike_data) * 100, 2)
		logger.debug(f'Calculated Position Delta: {position_delta}')
		if position_delta < 2 and position_delta > -2:
			# Delta is OK, check if price is sufficient
			new_position_price = self.calculate_position_price(legs, True)
			logger.debug(f'Calculated Position Price: {new_position_price:.2f}. Required price: {min_premium:.2f}')
			if new_position_price >= min_premium:
				# Store the legs which are above the required premium
				self.credit_legs = copy.deepcopy(legs)
				# Try to adjust
				return self.adjust_legs(position_delta, legs, strikes, option_strike_data, min_premium)
			else:
				# Legs Premium is below minimum premium --> So we adjusted to much
				# return the previous stored legs
				return self.credit_legs
		else:
			# Legs need to be adjusted
			return self.adjust_legs(position_delta, legs, strikes, option_strike_data, min_premium)

	def adjust_legs(self, position_delta: float, legs: List[Leg], strikes: list[float], option_strike_data, min_premium: float) -> List[Leg] | None:
		if position_delta > 0:
			# Short Put leg need to be further OTM
			for leg in legs:
				if leg.right == OptionRight.PUT and leg.action == OrderAction.SELL:
					# Find next OTM Put Strike
					next_strike = OptionHelper.get_next_strike(leg.strike, strikes, False)
					if next_strike is not None:
						leg.strike = next_strike
						logger.debug(f'Adjusted Put Leg to Strike: {next_strike}')

						# Long Put Leg adjustment
						long_leg = next((l for l in legs if l.right == OptionRight.PUT and l.action == OrderAction.BUY), None)
						if long_leg is not None:
							long_leg.strike = self.get_valid_strike(next_strike - self._template.wing, expiration_date=self._determined_expiration_date, higher=True)
							logger.debug(f'Adjusted Long Leg to Strike: {long_leg.strike}')
					else:
						logger.error('No further OTM Put Strike found, cannot adjust position delta.')
						return None
		else:
			# Short Call leg need to be further OTM
			for leg in legs:
				if leg.right == OptionRight.CALL and leg.action == OrderAction.SELL:
					# Find next OTM Call Strike
					next_strike = OptionHelper.get_next_strike(leg.strike, strikes, True)
					if next_strike is not None:
						leg.strike = next_strike
						logger.debug(f'Adjusted Call Leg to Strike: {next_strike}')

						# Long Leg adjustment
						long_leg = next((l for l in legs if l.right == OptionRight.CALL and l.action == OrderAction.BUY), None)
						if long_leg is not None:
							long_leg.strike = self.get_valid_strike(next_strike + self._template.wing, expiration_date=self._determined_expiration_date, higher=False)
							logger.debug(f'Adjusted Long Leg to Strike: {long_leg.strike}')
					else:
						logger.error('No further OTM Call Strike found, cannot adjust position delta.')
						return None
		return self.determine_order_legs(legs, strikes, option_strike_data, min_premium)

	def composeTakeProfitOrder(self, managedTrade: ManagedTrade, fillPrice: float) -> Order:
		"""
		Composes the take profit order based on the template and the given fill price
		"""
		super().composeTakeProfitOrder(managedTrade, fillPrice)
		logger.debug('Creating take profit order for template {}', self._template.name)
		takeProfitPrice = self._template.calculateTakeProfitPrice(fillPrice)
		logger.debug(f'Calculated take profit price: {takeProfitPrice}')
		tp_order_legs = copy.deepcopy(managedTrade.entryOrder.legs)
		self.invert_leg_actions(tp_order_legs)
		takeProfitOrder = Order(symbol=self._template.symbol, legs=tp_order_legs, action=OrderAction.SELL_TO_CLOSE, quantity=self._template.amount, type=OrderType.LIMIT, price=takeProfitPrice)
		return takeProfitOrder
	
	def composeStopLossOrder(self, managedTrade: ManagedTrade, fillPrice: float) -> Order:
		"""
		Composes the stop loss order based on the template and the given fill price
		"""
		super().composeStopLossOrder(managedTrade, fillPrice)
		logger.debug('Creating stop loss order for template {}', self._template.name)
		stopLossPrice = self._template.calculateStopLossPrice(fillPrice)
		sl_order_legs = copy.deepcopy(managedTrade.entryOrder.legs)
		self.invert_leg_actions(sl_order_legs)
		logger.debug(f'Calculated stop loss price: {stopLossPrice}')
		stopLossOrder = Order(symbol=self._template.symbol, legs=sl_order_legs, action=OrderAction.SELL_TO_CLOSE, quantity=self._template.amount, type=OrderType.STOP, price=stopLossPrice)
		return stopLossOrder


"""
Analysis API Endpoints

This module provides API endpoints for trade analysis and performance metrics.
"""

from datetime import datetime, timedelta
from enum import Enum
from typing import List, Optional

import pytz
from fastapi import APIRouter, Query
from loguru import logger
from pydantic import BaseModel
from sqlalchemy import or_, select

from optrabot.database import SessionLocal
from optrabot.models import Trade as TradeModel
from optrabot.tradestatus import TradeStatus

router = APIRouter(prefix='/api')


class AnalysisTimeRange(str, Enum):
    """Time range filter options for analysis"""
    TODAY = 'today'
    YESTERDAY = 'yesterday'
    THIS_WEEK = 'this_week'
    LAST_WEEK = 'last_week'
    THIS_MONTH = 'this_month'
    LAST_MONTH = 'last_month'
    THIS_YEAR = 'this_year'
    CUSTOM = 'custom'


class AnalysisTrade(BaseModel):
    """Trade information for analysis"""
    id: int
    strategy: str
    template_name: str
    trade_group: Optional[str]
    symbol: str
    status: str
    realized_pnl: Optional[float]
    fees: Optional[float]
    commission: Optional[float]
    entry_time: Optional[str]
    close_time: Optional[str]
    amount: int


class AnalysisData(BaseModel):
    """Analysis data response"""
    trades: List[AnalysisTrade]
    start_date: Optional[str]
    end_date: Optional[str]


def _get_analysis_time_range(
    time_range: AnalysisTimeRange,
    custom_start: Optional[str] = None,
    custom_end: Optional[str] = None
) -> tuple[Optional[datetime], Optional[datetime]]:
    """
    Get the start and end datetime for the given time range filter.
    Returns (start_date, end_date) tuple.
    """
    tz = pytz.timezone('US/Eastern')
    now = datetime.now(tz)
    
    if time_range == AnalysisTimeRange.TODAY:
        start = now.replace(hour=0, minute=0, second=0, microsecond=0)
        end = now
        
    elif time_range == AnalysisTimeRange.YESTERDAY:
        yesterday = now - timedelta(days=1)
        start = yesterday.replace(hour=0, minute=0, second=0, microsecond=0)
        end = yesterday.replace(hour=23, minute=59, second=59, microsecond=999999)
        
    elif time_range == AnalysisTimeRange.THIS_WEEK:
        # Start of current week (Monday)
        days_since_monday = now.weekday()
        week_start = now - timedelta(days=days_since_monday)
        start = week_start.replace(hour=0, minute=0, second=0, microsecond=0)
        end = now
        
    elif time_range == AnalysisTimeRange.LAST_WEEK:
        # Previous week (Monday to Sunday)
        days_since_monday = now.weekday()
        this_week_start = now - timedelta(days=days_since_monday)
        last_week_start = this_week_start - timedelta(days=7)
        last_week_end = this_week_start - timedelta(days=1)
        start = last_week_start.replace(hour=0, minute=0, second=0, microsecond=0)
        end = last_week_end.replace(hour=23, minute=59, second=59, microsecond=999999)
        
    elif time_range == AnalysisTimeRange.THIS_MONTH:
        start = now.replace(day=1, hour=0, minute=0, second=0, microsecond=0)
        end = now
        
    elif time_range == AnalysisTimeRange.LAST_MONTH:
        # First day of current month
        first_of_month = now.replace(day=1, hour=0, minute=0, second=0, microsecond=0)
        # Last day of previous month
        last_of_prev_month = first_of_month - timedelta(days=1)
        # First day of previous month
        first_of_prev_month = last_of_prev_month.replace(day=1)
        start = first_of_prev_month
        end = last_of_prev_month.replace(hour=23, minute=59, second=59, microsecond=999999)
        
    elif time_range == AnalysisTimeRange.THIS_YEAR:
        # First day of current year (January 1st)
        start = now.replace(month=1, day=1, hour=0, minute=0, second=0, microsecond=0)
        end = now
        
    elif time_range == AnalysisTimeRange.CUSTOM:
        # Parse custom dates
        start = None
        end = None
        if custom_start:
            try:
                start = datetime.fromisoformat(custom_start)
                if start.tzinfo is None:
                    start = tz.localize(start)
                # Ensure start is at beginning of day
                start = start.replace(hour=0, minute=0, second=0, microsecond=0)
            except ValueError:
                logger.warning(f'Invalid custom start date: {custom_start}')
        if custom_end:
            try:
                end = datetime.fromisoformat(custom_end)
                if end.tzinfo is None:
                    end = tz.localize(end)
                # Ensure end is at end of day (23:59:59.999999)
                end = end.replace(hour=23, minute=59, second=59, microsecond=999999)
            except ValueError:
                logger.warning(f'Invalid custom end date: {custom_end}')
    else:
        start = None
        end = None
    
    return start, end


@router.get('/analysis', response_model=AnalysisData)
async def get_analysis_data(
    time_range: AnalysisTimeRange = Query(AnalysisTimeRange.THIS_WEEK, description='Time range filter'),
    strategy: Optional[str] = Query(None, description='Filter by strategy name'),
    start_date: Optional[str] = Query(None, description='Custom start date (ISO format)'),
    end_date: Optional[str] = Query(None, description='Custom end date (ISO format)'),
    group_by_trade_group: bool = Query(False, description='Group trades by trade group')
) -> AnalysisData:
    """
    Get trade data for analysis.
    
    Returns closed trades within the specified time range with P&L data.
    """
    import optrabot.config as optrabotcfg

    # Get time range boundaries
    range_start, range_end = _get_analysis_time_range(time_range, start_date, end_date)
    logger.debug(f'Analysis request: time_range={time_range}, range_start={range_start}, range_end={range_end}')
    
    # Convert to UTC for database comparison (DB stores naive UTC timestamps)
    range_start_utc = range_start.astimezone(pytz.UTC).replace(tzinfo=None) if range_start else None
    range_end_utc = range_end.astimezone(pytz.UTC).replace(tzinfo=None) if range_end else None
    logger.debug(f'UTC range: start={range_start_utc}, end={range_end_utc}')
    
    # Get all templates for looking up template_group and strategy
    config: optrabotcfg.Config = optrabotcfg.appConfig
    all_templates = config.getTemplates() if config else []
    template_map = {t.name: t for t in all_templates}
    
    with SessionLocal() as session:
        # First, let's check what trades exist
        all_trades_query = select(TradeModel)
        all_trades = session.scalars(all_trades_query).all()
        logger.debug(f'Total trades in database: {len(all_trades)}')
        for t in all_trades[-5:]:  # Log last 5 trades
            logger.debug(f'  Trade {t.id}: status={t.status}, closeDate={t.closeDate}, openDate={t.openDate}')
        
        # Build query for closed/expired trades
        query = select(TradeModel).where(
            or_(
                TradeModel.status == TradeStatus.CLOSED,
                TradeModel.status == TradeStatus.EXPIRED
            )
        )
        
        # Apply time range filter
        # Use closeDate if available, otherwise fall back to openDate (for EXPIRED trades without closeDate)
        from sqlalchemy import func
        effective_date = func.coalesce(TradeModel.closeDate, TradeModel.openDate)
        
        if range_start_utc:
            query = query.where(effective_date >= range_start_utc)
        if range_end_utc:
            query = query.where(effective_date <= range_end_utc)
        
        # Order by effective date (closeDate or openDate)
        query = query.order_by(effective_date.asc())
        
        trades = session.scalars(query).all()
        logger.debug(f'Found {len(trades)} closed/expired trades in time range')
        
        result_trades: List[AnalysisTrade] = []
        
        for trade in trades:
            # Get template info for strategy name
            template = template_map.get(trade.template_name)
            strategy_name = template.strategy if template else (trade.strategy or trade.template_name)
            
            # Filter by strategy if specified
            if strategy and strategy_name != strategy:
                continue
            
            # Calculate fees and commission from transactions
            total_fees = 0.0
            total_commission = 0.0
            trade_amount = 0
            for tx in trade.transactions:
                if tx.commission:
                    total_commission += tx.commission
                if tx.fee:
                    total_fees += tx.fee
                if tx.contracts:
                    trade_amount = max(trade_amount, tx.contracts)
            
            result_trades.append(AnalysisTrade(
                id=trade.id,
                strategy=strategy_name,
                template_name=trade.template_name or '',
                trade_group=trade.trade_group_id,
                symbol=trade.symbol,
                status=trade.status,
                realized_pnl=round(trade.realizedPNL, 2) if trade.realizedPNL is not None else None,
                fees=round(total_fees, 2) if total_fees else None,
                commission=round(total_commission, 2) if total_commission else None,
                entry_time=trade.openDate.isoformat() if trade.openDate else None,
                # Use closeDate if available, otherwise fall back to openDate (for EXPIRED trades)
                close_time=(trade.closeDate or trade.openDate).isoformat() if (trade.closeDate or trade.openDate) else None,
                amount=trade_amount or 1,
            ))
        
        # If group_by_trade_group is True, aggregate trades by trade group
        if group_by_trade_group:
            grouped: dict[str, AnalysisTrade] = {}
            ungrouped: List[AnalysisTrade] = []
            
            for trade in result_trades:
                if trade.trade_group:
                    key = trade.trade_group
                    if key not in grouped:
                        grouped[key] = AnalysisTrade(
                            id=trade.id,  # Use first trade's ID
                            strategy=trade.strategy,
                            template_name=trade.template_name,
                            trade_group=trade.trade_group,
                            symbol=trade.symbol,
                            status=trade.status,
                            realized_pnl=trade.realized_pnl or 0,
                            fees=trade.fees or 0,
                            commission=trade.commission or 0,
                            entry_time=trade.entry_time,
                            close_time=trade.close_time,
                            amount=trade.amount,
                        )
                    else:
                        # Aggregate values
                        existing = grouped[key]
                        grouped[key] = AnalysisTrade(
                            id=existing.id,
                            strategy=existing.strategy,
                            template_name=existing.template_name,
                            trade_group=existing.trade_group,
                            symbol=existing.symbol,
                            status=existing.status,
                            realized_pnl=(existing.realized_pnl or 0) + (trade.realized_pnl or 0),
                            fees=(existing.fees or 0) + (trade.fees or 0),
                            commission=(existing.commission or 0) + (trade.commission or 0),
                            entry_time=existing.entry_time,  # Keep first entry time
                            close_time=trade.close_time,  # Use last close time
                            amount=existing.amount + trade.amount,
                        )
                else:
                    ungrouped.append(trade)
            
            result_trades = list(grouped.values()) + ungrouped
            # Re-sort by close time
            result_trades.sort(key=lambda t: t.close_time or '')
        
        return AnalysisData(
            trades=result_trades,
            start_date=range_start.isoformat() if range_start else None,
            end_date=range_end.isoformat() if range_end else None,
        )



"""
Module to define the config of the application
"""
import os
from collections import UserDict
from typing import List, OrderedDict

from InquirerPy import inquirer
from InquirerPy.base.control import Choice
from InquirerPy.validator import EmptyInputValidator, NumberValidator
from loguru import logger
from ruyaml import YAML
from ruyaml.scanner import ScannerError

from optrabot.flowengine.flowconfig import (Flow, FlowAction, FlowActionType,
                                            FlowEventConfig,
                                            ProcessTemplateAction,
                                            SendNotificationAction)
from optrabot.flowengine.flowevent import FlowEventType
from optrabot.tradetemplate.earlyexittrigger import EarlyExitTrigger
from optrabot.tradetemplate.templatedata import LongStrikeData, ShortStrikeData
from optrabot.tradetemplate.templatefactory import (IronCondor, IronFly,
                                                    PutSpread, Template,
                                                    TemplateFactory,
                                                    TemplateType)
from optrabot.tradetemplate.templatetrigger import TemplateTrigger

configFileName = "config.yaml"
defaultHubHost = 'wss://app.optrabot.io'

# Global variable to store the current configuration
appConfig = None

class Config(UserDict):
	def __init__(self, config_path=configFileName):
		self.config_path = os.path.expanduser(config_path)
		self._templates = []
		self._flows = []
		self.loaded = False
		self.load()

	def load(self):
		"""
        Loads configuration from configuration YAML file.
        """
		logger.debug("Try loading config file...")
		try:
			# Open with UTF-8 encoding explicitly to support emojis and special characters on Windows
			with open(os.path.expanduser(self.config_path), 'r', encoding='utf-8') as f:
				try:
					self.data = YAML().load(f)
					logger.debug("Config loaded successfully.")
					self.loaded = True
				except ScannerError as e:
					logger.error("Error parsing yaml of configuration file '{}' :{}", e.problem_mark, e.problem)
				except UnicodeDecodeError as e:
					logger.error("Error decoding configuration file '{}': {}. Make sure the file is saved as UTF-8.", self.config_path, str(e))
				except Exception as e:
					logger.error("Error loading configuration file '{}': {}", self.config_path, str(e))
		except FileNotFoundError:
			logger.error(
				 "Error opening configuration file {}".format(self.config_path)
			)
			self.createDefaultConfig()

	def createDefaultConfig(self):
		""" Create a default configuration 
		"""
		logger.info('Using a default configuration.')
		defaultDoc = 'general:\n'
		defaultDoc += '  instanceid: testagent\n'
		self.data = YAML().load(defaultDoc)

	def get(self, key):
		"""
		Fetch the configuration value for the specified key. If there are nested dictionaries, a dot
		notation can be used.

		So if the configuration contents are:

		self.data = {
			'first': {
				'second': 'value'
			},
		}

        self.data.get('first.second') == 'value'

		Arguments:
        	key(str): Configuration key to fetch
		"""
		keys = key.split('.')
		value = self.data.copy()

		for key in keys:
			value = value[key]
			if value == None:
				raise KeyError

		return value
	
	def save(self):
		"""
		Saves configuration in the configuration YAML file.
		"""
		with open(os.path.expanduser(self.config_path), 'w+') as f:
			yaml = YAML()
			yaml.default_flow_style = False
			yaml.dump(self.data, f)

	def logConfigurationData(self):
		"""
		Write Configuration data into the optrabot log file in Debug Mode
		"""
		logger.debug('Using following configuration data...')
		for key, value in self.data.items():
			logger.debug('Category {}:', key)
			if value == None:
				continue
			children = OrderedDict(value)
			for subkey, subvalue in children.items():
				if isinstance(subvalue, dict):
					logger.debug('Sub Category {}:', subkey)
					try:
						subChildren = OrderedDict(subvalue)
						for subsubkey, subsubvalue in subChildren.items():
							if isinstance(subsubvalue, dict):
								logger.debug('	{}', subsubkey)
								try:
									subsubChildren = OrderedDict(subsubvalue)
									for subsubsubkey, subsubsubvalue in subsubChildren.items():
										logger.debug('       	{}: {}', subsubsubkey, subsubsubvalue)
								except ValueError:
									continue
							else:
								logger.debug('       {}: {}', subsubkey, subsubvalue)
					except ValueError:
						continue
				else:
					logger.debug('   {}: {}', subkey, subvalue)

	def readTemplates(self):
		self._templates = []
		try:
			for key, data in self.get('templates').items():
				logger.debug('Reading Template: {}', key)
				template = TemplateFactory.createTemplate(name=key,data=data)
				if template != None:
					if template.validate():
						template_state = '' if template.is_enabled() else '(disabled)'
						logger.info(f'Loaded Trade Template: {template.name} {template_state}')
						self._templates.append(template)
					else:
						logger.error(f'Template {template.name} is not valid. Please check the configuration!')
		except KeyError as exc:
			logger.error('Missing key: {} --> No templates are configured!', exc)

	def getTemplates(self) -> List[Template]:
		""" Returns the configured templates
		"""
		return self._templates
	
	def get_template_by_name(self, template_name: str) -> Template | None:
		"""
		Find a template by its name.
		
		Args:
			template_name: Name of the template to find
			
		Returns:
			Template object if found, None otherwise
			
		Example:
			template = config.get_template_by_name("IC_SPX_0DTE")
		"""
		if not template_name:
			return None
		
		for template in self._templates:
			if template.name == template_name:
				return template
		
		return None
	
	def readFlows(self):
		"""
		Read flow configurations from the config file
		"""
		self._flows = []
		try:
			flows_config = self.get('flows')
			if flows_config is None:
				logger.debug('No flows configured')
				return
			
			# Handle both list and dict format
			if isinstance(flows_config, list):
				# List format: flows: - flow_id: {...}
				for flow_item in flows_config:
					if not isinstance(flow_item, dict):
						logger.warning(f'Invalid flow configuration item: {flow_item}')
						continue
					for flow_id, flow_data in flow_item.items():
						flow = self._parse_flow(flow_id, flow_data)
						if flow:
							self._flows.append(flow)
			elif isinstance(flows_config, dict):
				# Dict format: flows: flow_id: {...}
				for flow_id, flow_data in flows_config.items():
					flow = self._parse_flow(flow_id, flow_data)
					if flow:
						self._flows.append(flow)
			else:
				logger.error(f'Invalid flows configuration format: {type(flows_config)}')
				return
			
			logger.info(f'Loaded {len(self._flows)} flow configuration(s)')
			
		except KeyError:
			logger.debug('No flows section in configuration file')
	
	def _parse_flow(self, flow_id: str, flow_data: dict) -> Flow:
		"""
		Parse a single flow configuration
		
		Args:
			flow_id: Flow identifier
			flow_data: Flow configuration data
			
		Returns:
			Flow object or None if parsing fails
		"""
		try:
			# Extract basic flow properties
			name = flow_data.get('name')
			enabled = flow_data.get('enabled', True)
			
			# Parse event configuration
			event_config = None
			event_data = flow_data.get('event')
			if event_data:
				try:
					event_type = FlowEventType(event_data.get('type'))
					template_name = event_data.get('template')
					if not template_name:
						logger.error(f'Flow "{flow_id}": Missing required "template" in event configuration')
						return None
					
					# Validate that the template exists and is enabled
					template_found = False
					template_enabled = False
					for template in self._templates:
						if template.name == template_name:
							template_found = True
							template_enabled = template.is_enabled()
							break
					
					if not template_found:
						logger.warning(f'Flow "{flow_id}": Referenced template "{template_name}" does not exist - Flow will be disabled')
						enabled = False
					elif not template_enabled:
						logger.warning(f'Flow "{flow_id}": Referenced template "{template_name}" is disabled - Flow will be disabled')
						enabled = False
					
					event_config = FlowEventConfig(type=event_type, template=template_name)
				except ValueError as e:
					logger.error(f'Flow "{flow_id}": Invalid event type "{event_data.get("type")}"')
					return None
			else:
				logger.error(f'Flow "{flow_id}": Missing event configuration')
				return None
			
			# Parse actions
			actions = []
			actions_data = flow_data.get('actions', [])
			for action_item in actions_data:
				if not isinstance(action_item, dict):
					logger.warning(f'Flow "{flow_id}": Invalid action item: {action_item}')
					continue
				
				for action_type, action_config in action_item.items():
					if action_type == 'send_notification':
						message = action_config.get('message', '')
						notification_type = action_config.get('type', 'INFO')
						action = FlowAction(
							action_type=FlowActionType.SEND_NOTIFICATION,
							action_config=SendNotificationAction(message=message, type=notification_type)
						)
						actions.append(action)
					elif action_type == 'process_template':
						template_name = action_config.get('template')
						amount = action_config.get('amount')
						premium = action_config.get('premium')
						expiration = action_config.get('expiration')  # Optional
						time = action_config.get('time')  # Optional
						
						if not template_name:
							logger.error(f'Flow "{flow_id}": Missing template in process_template action')
							continue
						if amount is None:
							logger.error(f'Flow "{flow_id}": Missing amount in process_template action')
							continue
						if premium is None:
							logger.error(f'Flow "{flow_id}": Missing premium in process_template action')
							continue
						
						# Parse and validate time format if provided
						parsed_time = None
						if time is not None:
							from optrabot.tradetemplate.templatetrigger import TemplateTrigger
							try:
								# Parse the time using the same logic as TemplateTrigger
								trigger = TemplateTrigger({'type': 'time', 'value': time})
								parsed_time = trigger.parseTimeWithTimezone(time)
								if parsed_time is None:
									logger.error(f'Flow "{flow_id}": Invalid time format "{time}" - Expected format: HH:MM <Timezone> (e.g. "15:00 EST")')
									logger.error(f'Flow "{flow_id}": Flow will be disabled due to invalid time configuration')
									enabled = False
									continue
								logger.debug(f'Flow "{flow_id}": Successfully parsed time parameter: {time} -> {parsed_time}')
							except Exception as e:
								logger.error(f'Flow "{flow_id}": Error parsing time "{time}": {str(e)}')
								logger.error(f'Flow "{flow_id}": Flow will be disabled due to invalid time configuration')
								enabled = False
								continue
						
						# Validate that the template exists and is enabled
						template_found = False
						template_enabled = False
						for template in self._templates:
							if template.name == template_name:
								template_found = True
								template_enabled = template.is_enabled()
								break
						
						if not template_found:
							logger.warning(f'Flow "{flow_id}": Action references template "{template_name}" that does not exist - Flow will be disabled')
							enabled = False
							continue
						elif not template_enabled:
							logger.warning(f'Flow "{flow_id}": Action references template "{template_name}" that is disabled - Flow will be disabled')
							enabled = False
							continue
						
						action = FlowAction(
							action_type=FlowActionType.PROCESS_TEMPLATE,
							action_config=ProcessTemplateAction(
								template=template_name,
								amount=amount,
								premium=premium,
								expiration=expiration,
								time=parsed_time  # Store the parsed datetime object
							)
						)
						actions.append(action)
					else:
						logger.warning(f'Flow "{flow_id}": Unknown action type "{action_type}"')
			
			if not actions:
				logger.warning(f'Flow "{flow_id}": No valid actions configured')
			
			# Create flow object
			flow = Flow(
				id=flow_id,
				name=name,
				enabled=enabled,
				event=event_config,
				actions=actions
			)
			
			# Log flow loading with name or ID and enabled status
			state = 'enabled' if enabled else 'disabled'
			display_name = name if name else flow_id
			logger.info(f'Loaded Flow: {display_name} ({state})')
			return flow
			
		except Exception as e:
			logger.error(f'Error parsing flow "{flow_id}": {e}')
			return None
	
	def getFlows(self) -> List[Flow]:
		"""
		Returns the configured flows
		"""
		return self._flows
	
	def get_spread_monitoring_config(self):
		"""
		Returns the spread monitoring configuration.
		
		Reads the spread_monitoring section from config.yaml and returns
		a SpreadMonitoringConfig object with the configured values or defaults.
		
		Returns:
			SpreadMonitoringConfig: Configuration object for spread monitoring
			
		Example config.yaml:
			general:
			  spread_monitoring:
			    mode: active_strikes
			    multiplier: 2.0
			    min_ask_threshold: 0.30
		"""
		from optrabot.broker.spreadmonitoring import SpreadMonitoringConfig
		try:
			spread_config = self.data.get('general', {}).get('spread_monitoring', {})
		except (KeyError, AttributeError):
			spread_config = {}
		return SpreadMonitoringConfig.from_config_dict(spread_config)
	
	def getInstanceId(self) -> str:
		""" Returns the configured OptraBot Agent Id / Instance Id
		"""
		instanceId = ''
		try:
			instanceId = self.get('general.instanceid')
		except KeyError as exc:
			instanceId = self.get('general.agentid')
		return instanceId

def captureTemplateData(template: Template):
	template.strategy = inquirer.text(message="Strategy:", default=template.strategy).execute()
	template.adjustmentStep = float(inquirer.number(message="Increment for limit price adjustments ($):",default=template.adjustmentStep, float_allowed=True, validate=EmptyInputValidator()).execute())
	#if template._type == TemplateType.IronFly:
	#	template.wing = int(inquirer.number(message="Wing Size (points):", default=template.wing, min_allowed=10).execute())
	if template._type == TemplateType.PutSpread:
		short_strike_data = template.getShortStrikeData()
		short_strike_data.offset = int(inquirer.number(message="Offset for the short strike ($):", default=short_strike_data.offset).execute())
		template.setShortStikeData(short_strike_data)
		long_strike_data = template.getLongStrikeData()
		long_strike_data.width = int(inquirer.number(message="Width of the Spread (points):", default=long_strike_data.width, min_allowed=1).execute())
		template.setLongStikeData(long_strike_data)
	template.account = inquirer.text(message="Account number to be used for trading:",default=template.account, validate=EmptyInputValidator()).execute()
	template.amount = int(inquirer.number(message="Number of contracts to be traded:", default=template.amount, min_allowed=1).execute())
	if template.takeProfit is not None:
		try:
			template.takeProfit = int(inquirer.number(message="Percentage of captured premium as profit target (%):", default=template.takeProfit, min_allowed=1, max_allowed=100).execute())
		except ValueError:
			template.takeProfit = None
	if template.stopLoss is not None:
		template.stopLoss = int(inquirer.number(message="Percentage of captured premium as stop loss level (%):", default=template.stopLoss, min_allowed=1).execute())
	if template._type == TemplateType.IronCondor:
		if template.wing is not None:
			template.wing = int(inquirer.number(message="Wing Size (points):", default=template.wing, min_allowed=5).execute())
		if template.premium is not None:
			template.premium = float(inquirer.number(message="Expected premium for an Iron Condor trade ($):",default=template.premium, float_allowed=True, validate=EmptyInputValidator()).execute())
	
	if template._type == TemplateType.IronFly:
		template.minPremium = float(inquirer.number(message="Minimum premium for an Iron Fly trade ($):",default=template.minPremium, float_allowed=True, validate=EmptyInputValidator()).execute())
	try:
		template.maxOpenTrades = int(inquirer.number(message="Maximum number of parallel open positions:", default=template.maxOpenTrades, min_allowed=1).execute())
	except ValueError:
		template.maxOpenTrades = 0

def ensureInitialConfig() -> bool:
	configPath = os.path.expanduser(configFileName)
	if os.path.exists(configPath):
		return True
	print("No config.yaml found. Let's answer some questions and generate the required configuration file.")
	configOK = False
	try:
		confAPIKey = inquirer.text(message="What's your OptraBot API Key:").execute()
		confInstanceId = inquirer.text(message="Give your OptraBot Instance an Id:", default="optrabot").execute()
		confWebPort = inquirer.number(message="Port number on which the OptraBot UI will be accessible:", default=8080).execute()
		confTWSHost = inquirer.text(message="Hostname of your TWS/IB Gateway machine:",default="127.0.0.1").execute()
		confTWSPort = inquirer.number(message="Port number of your TWS/IB Gateway:",default=7496, validate=NumberValidator()).execute()
		confTWSClientID = inquirer.number(message="Client ID to be used for TWS/IB Gateway connection:",default=21, validate=NumberValidator()).execute()
		confTWSMarketData = inquirer.select(
			message="Select a market data type:",
			choices=[
	            "Live",
	            "Delayed",
	            #Choice(value=None, name="Exit"),
	        ],
	        default="Live",
		).execute()

		templatesDict = {}
		addTemplate = True
		templCount = 0
		while templCount < 3:
			templCount += 1
			if templCount == 1:
				templateDefault = IronCondor('0DTEIIC')
				templateDefault.strategy = '0DTE Income Iron Condor'
				templateDefault._trigger = TemplateTrigger({'type': 'time', 'value': '09:32 EST'})
				templateDefault.adjustmentStep = 0.05
				templateDefault.wing = 15
				templateDefault.takeProfit = None
				templateDefault.stopLoss = None
				templateDefault.account = ''
				templateDefault.amount = 1
				templateDefault.set_enabled(True)
				templateDefault.premium = 0.6
				templateDefault.vix_max = 25
				templateDefault.set_early_exit(EarlyExitTrigger({'type': 'breakeven'}))
			elif templCount == 2 :
				templateDefault = PutSpread('0DTEMagickTrend10')
				templateDefault.strategy = 'Magick Trend 10'
				templateDefault._trigger = TemplateTrigger({'type': 'external', 'value': '0DTEMagickTrend10', 'excludetimefrom': '11:30 EST', 'excludetimeto': '12:00 EST'})
				templateDefault.adjustmentStep = 0.05
				short_strike_data = ShortStrikeData()
				short_strike_data.offset = -5
				templateDefault.setShortStikeData(short_strike_data)
				long_strike_data = LongStrikeData()
				long_strike_data.width = 70
				templateDefault.setLongStikeData(long_strike_data)
				templateDefault.account = ''
				templateDefault.amount = 1
				templateDefault.takeProfit = 50
				templateDefault.stopLoss = 200
				templateDefault.maxOpenTrades = None
				templateDefault.set_enabled(True)
				# templateDefault.minPremium = 14
			elif templCount == 3:
				templateDefault = PutSpread('0DTEMagickTrend30')
				templateDefault.strategy = 'Magick Trend 30'
				templateDefault._trigger = TemplateTrigger({'type': 'external', 'value': '0DTEMagickTrend30', 'timefrom': '11:00 EST', 'timeto': '15:00 EST'})
				templateDefault.adjustmentStep = 0.05
				short_strike_data = ShortStrikeData()
				short_strike_data.offset = -5
				templateDefault.setShortStikeData(short_strike_data)
				long_strike_data = LongStrikeData()
				long_strike_data.width = 70
				templateDefault.setLongStikeData(long_strike_data)
				templateDefault.account = ''
				templateDefault.amount = 1
				templateDefault.stopLoss = 200
				templateDefault.maxOpenTrades = None
				templateDefault.set_enabled(True)
			# elif templCount == 2:
			# 	templateDefault = IronFly('0DTELunchIF')
			# 	templateDefault.strategy = '0DTE Lunch Iron Fly'
			# 	templateDefault._trigger = TemplateTrigger({'type': 'external', 'value': '0DTELunchIF'})
			# 	templateDefault.adjustmentStep = 0.1
			# 	templateDefault.wing = 70
			# 	templateDefault.account = ''
			# 	templateDefault.amount = 1
			# 	templateDefault.takeProfit = 10
			# 	templateDefault.stopLoss = 20
			# 	templateDefault.minPremium = 0
			# elif templCount == 3:
			# 	templateDefault = IronFly('0DTEFOMCFly')
			# 	templateDefault.strategy = '0DTE FOMC Fly'
			# 	templateDefault._trigger = TemplateTrigger({'type': 'external', 'value': '0DTEFOMCFly'})
			# 	templateDefault.adjustmentStep = 0.1
			# 	templateDefault.wing = 100
			# 	templateDefault.account = ''
			# 	templateDefault.amount = 1
			# 	templateDefault.takeProfit = 10
			# 	templateDefault.stopLoss = 100
			# 	templateDefault.minPremium = 0
			else:
				break
			addTemplate = inquirer.confirm(message="Add a trade template '" +  templateDefault.name + "'", default=True).execute()
			if not addTemplate:
				continue
			captureTemplateData(templateDefault)
			templatesDict.update({templateDefault.name:templateDefault.toDict()})

		# confTradingAccount = inquirer.text(message="Account number to be used for trading:",default="").execute()
		# confTradingContracts = inquirer.number(message="Number of IronFly contracts to be traded:", min_allowed=1).execute()
		# confTradingPriceIncement = inquirer.number(message="Increment for limit price adjustments ($):",default=0.1, float_allowed=True, validate=EmptyInputValidator()).execute()
		# confTradingMinimumPremium = inquirer.number(message="Minimum premium for an Iron Fly trade ($):",default=14, float_allowed=True, validate=EmptyInputValidator()).execute()
		# confTradingTakeProfit = inquirer.number(message="Percentage of captured premium as profit target (%):", default=8, min_allowed=1, max_allowed=100).execute()
		# confTradingStopLoss = inquirer.number(message="Percentage of captured premium as stop loss level (%):", default=16, min_allowed=1).execute()
		# confStopLossAdjust = inquirer.confirm(message="Automatic stop loss adjustment?",default=False).execute()
		# if confStopLossAdjust:
		# 	confSLATrigger = inquirer.number(message="Percentage of profit that triggers the stop loss adjustment (%):", default=10, min_allowed=1, validate=EmptyInputValidator()).execute()
		# 	confSLAStop = inquirer.number(message="Adjusted stop 'postive=below entry; negative=above entry' (%):", default=0, validate=EmptyInputValidator()).execute()
		# 	confSLAOffset = inquirer.number(message="Offset for the adjusted stop 'negtive=away from market; positive=closer to market' ($):", default=-0.2, float_allowed=True,validate=EmptyInputValidator()).execute()
		confirm = inquirer.confirm(message="Confirm?", default=True).execute()
	except KeyboardInterrupt as exc:
		print('Configuration has been aborted!')
		return False

	if not confirm:
		print("Configuration assistant abortet!")
		return

	data = dict(general=dict(port=int(confWebPort), apikey=confAPIKey, instanceid=confInstanceId, hub=defaultHubHost),
				tws=dict(host=confTWSHost, port=int(confTWSPort), clientid=int(confTWSClientID), marketdata=confTWSMarketData),templates=templatesDict)
	#if confStopLossAdjust:
	#	twsItems = dict(data["tws"].items())
	#	twsItems["adjuststop"] = dict(trigger=int(confSLATrigger), stop=int(confSLAStop), offset=float(confSLAOffset))
	#	data["tws"] = twsItems
		  
	configYAML = yaml=YAML()
	try:
		with open(configFileName, 'w') as configFile:
			configYAML.dump(data, configFile )
		print("Config file " + configFileName + " has been generated based on your answers.\nYou may modify the configuration file manually if required.")
		configOK = True
	except IOError as exc:
		print("Error generating the config file:", configFileName)
		print("I/O error({0}): {1}".format(exc.errno, exc.strerror))
		return
	
	return configOK


import asyncio
import datetime as dt
import re
import ssl
from dataclasses import dataclass
from datetime import date, timedelta
from decimal import Decimal
from typing import Dict, List

import certifi
from apscheduler.schedulers.asyncio import AsyncIOScheduler
from httpx import ConnectError
from loguru import logger
from pydantic import ValidationError
from tastytrade import Account, AlertStreamer, DXLinkStreamer, Session
from tastytrade.account import Transaction
from tastytrade.dxfeed import Candle, Greeks, Quote
from tastytrade.instruments import (NestedOptionChain, Option, OptionType,
                                    Strike)
from tastytrade.order import (NewComplexOrder, NewOrder, OrderAction,
                              OrderStatus, OrderTimeInForce, OrderType,
                              PlacedOrder)
from tastytrade.utils import TastytradeError
from websockets.exceptions import ConnectionClosedOK

import optrabot.config as optrabotcfg
import optrabot.symbolinfo as symbolInfo
from optrabot.broker.brokerconnector import BrokerConnector, SymbolData
from optrabot.broker.optionpricedata import (OptionStrikeData,
                                             OptionStrikePriceData)
from optrabot.broker.order import Execution
from optrabot.broker.order import Leg as GenericOrderLeg
from optrabot.broker.order import OptionRight
from optrabot.broker.order import Order as GenericOrder
from optrabot.broker.order import OrderAction as GenericOrderAction
from optrabot.broker.order import OrderStatus as GenericOrderStatus
from optrabot.broker.order import PriceEffect
from optrabot.exceptions.orderexceptions import (PlaceOrderException,
                                                 PrepareOrderException)
from optrabot.managedtrade import ManagedTrade
from optrabot.models import Account as ModelAccount
from optrabot.optionhelper import OptionHelper
from optrabot.tradetemplate.templatefactory import Template


@dataclass
class TastySymbolData(SymbolData):
	def __init__(self) -> None:
		super().__init__()
		self.tastySymbol: str = None
		self.chain: NestedOptionChain = None

class TastytradeConnector(BrokerConnector):
	task_listen_quotes: asyncio.Task = None

	def __init__(self) -> None:
		super().__init__()
		self._client_secret = ''
		self._refresh_token = ''
		self._sandbox = False
		self._initialize()
		self.id = 'TASTY'
		self.broker = 'TASTY'
		self._orders: List[GenericOrder] = []
		self._replacedOrders: List[PlacedOrder] = []
		self._processed_transaction_ids: List[int] = []
		self._is_disconnecting = False
		self._is_reconnecting_alert_streamer = False  # Guard against multiple simultaneous reconnects
		self._session: Session = None
		self._streamer: DXLinkStreamer = None
		self._alert_streamer: AlertStreamer = None
		self._quote_symbols = []
		self._candle_symbols = []
		self._greeks_symbols = []
		self._tasty_accounts: List[Account] = []
		self._symbolReverseLookup: Dict[str, str] = {}		# maps tastytrade symbol to generic symbol

		self.task_listen_quotes = None
		self.task_listen_accounts = None
		self.task_listen_greeks = None
		self.task_listen_candle = None

		self._backgroundScheduler = AsyncIOScheduler()

	def _initialize(self) -> None:
		"""
		Initialize the Tastytrade connector from the configuration
		"""
		if not optrabotcfg.appConfig:
			return
		
		config :optrabotcfg.Config = optrabotcfg.appConfig
		try:
			config.get('broker.tastytrade')
		except KeyError:
			logger.debug('No Tastytrade connection configured')
			return
		
		try:
			self._client_secret = config.get('broker.tastytrade.client_secret')
		except KeyError:
			logger.error('Tastytrade client_secret not configured')
			return
		
		try:
			self._refresh_token = config.get('broker.tastytrade.refresh_token')
		except KeyError:
			logger.error('Tastytrade refresh_token not configured')
			return
		
		try:
			self._sandbox = config.get('broker.tastytrade.sandbox')
		except KeyError:
			pass
		self._initialized = True

	async def cancel_order(self, order: GenericOrder) -> None:
		""" 
		Cancels the given order
		"""
		await super().cancel_order(order)
		tasty_order: PlacedOrder = order.brokerSpecific['order']
		account: Account = order.brokerSpecific['account']
		logger.debug(f'Cancelling order {tasty_order.id}')
		try:
			account.delete_order(self._session, str(tasty_order.id))
		except TastytradeError as tastyErr:
			logger.error(f'Error cancelling order {tasty_order.id}: {tastyErr}')
			raise

	async def connect(self) -> None:
		await super().connect()
		self._is_disconnecting = False
		self._is_reconnecting_alert_streamer = False  # Reset reconnect guard on new connection
		try:
			#self._session = Session(self._username, self._password, is_test=self._sandbox)
			self._session = Session(provider_secret=self._client_secret, refresh_token=self._refresh_token, is_test=self._sandbox)
			await self.set_trading_enabled(True, 'Broker connected')
			
			# I Sandbox mode, there are SPX optins only
			if self._sandbox:
				spx_symbol_information = symbolInfo.symbol_infos['SPX']
				spx_symbol_information.trading_class = 'SPX'
				spx_symbol_information.option_symbol_suffix = ''

			# OTB-xxx: Verify session is ready by attempting to retrieve accounts
			# This prevents race conditions where connectedEvent is emitted before
			# the session is fully initialized and ready to handle API calls.
			# If successful, accounts are cached for subsequent getAccounts() calls.
			max_retries = 3
			session_ready = False
			for attempt in range(max_retries):
				try:
					self._tasty_accounts = Account.get(self._session)
					if len(self._tasty_accounts) > 0:
						logger.debug(f'Session verified - {len(self._tasty_accounts)} account(s) accessible')
						session_ready = True
						break
					else:
						logger.debug(f'Session not ready - no accounts found (attempt {attempt + 1}/{max_retries})')
				except Exception as e:
					logger.debug(f'Session verification failed (attempt {attempt + 1}/{max_retries}): {e}')

				if attempt < max_retries - 1:
					await asyncio.sleep(0.2)  # Brief delay between retries
			
			if not session_ready:
				logger.warning('Session created but verification failed after 3 attempts - proceeding with caution')
			
			self._backgroundScheduler.start()
			self._backgroundScheduler.add_job(self._perform_periodic_updates, 'interval', seconds=10, misfire_grace_time=None)

			self._emitConnectedEvent()
		except TastytradeError as tastyErr:
			logger.error('Failed to connect to Tastytrade: {}', tastyErr)
			self._emitConnectFailedEvent()
		except ConnectError:
			logger.debug('Failed to connect to Tastytrade - network connection error')
			self._emitConnectFailedEvent()

	async def _disconnect_internal(self) -> None:
		"""
		Perform the operations for disconnecting from Tastytrade
		"""
		# Note: _is_disconnecting should already be set to True before calling this method
		logger.debug('Starting _disconnect_internal')
		await self.set_trading_enabled(False, 'Broker disconnected')

		# FIRST: Close streamers gracefully to stop their internal _connect tasks
		# This prevents the RecursionError from nested task cancellation
		if self._streamer:
			try:
				logger.debug('Closing DXLink streamer')
				await asyncio.wait_for(self._streamer.close(), timeout=2.0)
				logger.debug('DXLink streamer closed')
			except asyncio.TimeoutError:
				logger.warning('Timeout closing DXLink streamer')
			except Exception as e:
				logger.debug(f'Error closing DXLink streamer: {e}')
		
		if self._alert_streamer:
			try:
				logger.debug('Closing Alert streamer')
				await asyncio.wait_for(self._alert_streamer.close(), timeout=2.0)
				logger.debug('Alert streamer closed')
			except asyncio.TimeoutError:
				logger.warning('Timeout closing Alert streamer')
			except Exception as e:
				logger.debug(f'Error closing Alert streamer: {e}')

		# SECOND: Now cancel our listener tasks (which should already be stopping)
		tasks_to_cancel = [
			task for task in [
				self.task_listen_quotes,
				self.task_listen_greeks,
				self.task_listen_candle,
				self.task_listen_accounts
			] if task and not task.done()
		]

		if tasks_to_cancel:
			logger.debug(f'Cancelling {len(tasks_to_cancel)} listener tasks')
			for task in tasks_to_cancel:
				task.cancel()
			
			# Give tasks a moment to process cancellation
			await asyncio.sleep(0.1)

		# THIRD: Clear streamer references
		logger.debug('Clearing streamer references')
		self._streamer = None
		self._alert_streamer = None
		logger.debug('Streamer references cleared')

		# FOURTH: Let event loop clean up cancelled tasks
		if tasks_to_cancel:
			logger.debug('Listener tasks cancelled - will be cleaned up by event loop')
		
		# Give event loop time to process cancellations
		await asyncio.sleep(0.1)

		# Clear task references
		self.task_listen_quotes = None
		self.task_listen_greeks = None
		self.task_listen_candle = None
		self.task_listen_accounts = None

		# Close session
		self._session = None
		
		# Clear cached account data
		self._tasty_accounts = []
		
		self._emitDisconnectedEvent()

	async def disconnect(self):
		"""
		Disconnect from Tastytrade broker.
		Sets _is_disconnecting flag BEFORE starting disconnect operations to prevent
		reconnect attempts from disconnect callbacks.
		"""
		await super().disconnect()
		# Set flag BEFORE calling _disconnect_internal to prevent reconnect attempts
		self._is_disconnecting = True
		await self._disconnect_internal()

	def getAccounts(self) -> List[ModelAccount]:
		"""
		Returns the Tastytrade accounts and starts the account update task for listening to updates on orders.
		Uses cached account data from connect() if available.
		"""
		if len(self._managedAccounts) == 0 and self.isConnected():
			# If accounts were not already loaded in connect(), load them now
			if len(self._tasty_accounts) == 0:
				self._tasty_accounts = Account.get(self._session)
			
			for tastyAccount in self._tasty_accounts:
				account = ModelAccount(id = tastyAccount.account_number, name = tastyAccount.nickname, broker = self.broker, pdt = not tastyAccount.day_trader_status)
				self._managedAccounts.append(account)

			# Store the task reference so we can cancel it during disconnect
			self.task_listen_accounts = asyncio.create_task(self._request_account_updates())
		return self._managedAccounts
	
	def isConnected(self) -> bool:
		return self._session is not None
		
	async def prepareOrder(self, order: GenericOrder, need_valid_price_data: bool = True) -> None:
		"""
		Prepares the given order for execution

		Args:
			order: The order to prepare
			need_valid_price_data: If True, validates price data availability and freshness.
								   If False, creates order legs without price validation (e.g., for trade recovery)

		Raises:
			PrepareOrderException: If the order preparation fails with a specific reason.
		"""
		symbolData = self._symbolData[order.symbol]
		comboLegs: list[GenericOrderLeg] = []
		for leg in order.legs:
			optionInstrument: Option = None
			
			if need_valid_price_data:
				# Normal order flow: use price data with validation
				try:
					optionPriceData = symbolData.optionPriceData[leg.expiration]
				except KeyError as keyErr:
					raise PrepareOrderException(f'No option price data for expiration date {leg.expiration} available!', order)

				try:
					priceData: OptionStrikePriceData = optionPriceData.strikeData[leg.strike]
					if not priceData.is_outdated():
						if leg.right == OptionRight.CALL:
							leg.askPrice = float(priceData.callAsk)
							if leg.askPrice == None:
								leg.askPrice = 0
							leg.bidPrice = float(priceData.callBid)
							if leg.bidPrice == None:
								leg.bidPrice = 0
							optionInstrument = Option.get(self._session, priceData.brokerSpecific['call_option'])
						elif leg.right == OptionRight.PUT:
							leg.askPrice = float(priceData.putAsk)
							if leg.askPrice == None:
								leg.askPrice = 0
							leg.bidPrice = float(priceData.putBid)
							if leg.bidPrice == None:
								leg.bidPrice = 0
							optionInstrument = Option.get(self._session, priceData.brokerSpecific['put_option'])
					else:
						raise PrepareOrderException(f'Price data for strike {leg.strike} is outdated or not available!', order)

				except KeyError as keyErr:
					# No data for strike available
					raise PrepareOrderException(f'No option price data for strike {leg.strike} available!', order)
				except Exception as excp:
					raise PrepareOrderException(f'Error preparing order: {excp}', order)
			else:
				# Trade recovery flow: create option instrument without price data validation
				try:
					# Use the cached option chain from symbolData
					if symbolData.chain is None:
						raise PrepareOrderException(
							f'No option chain available for {order.symbol} - symbol data may not be initialized', 
							order
						)
					
					# Find the specific option in the cached chain
					for expiration in symbolData.chain.expirations:
						if expiration.expiration_date == leg.expiration:
							for strike in expiration.strikes:
								if abs(float(strike.strike_price) - leg.strike) < 0.01:  # Float comparison tolerance
									if leg.right == OptionRight.CALL and strike.call is not None:
										optionInstrument = Option.get(self._session, strike.call)
										leg.askPrice = 0
										leg.bidPrice = 0
										break
									elif leg.right == OptionRight.PUT and strike.put is not None:
										optionInstrument = Option.get(self._session, strike.put)
										leg.askPrice = 0
										leg.bidPrice = 0
										break
							if optionInstrument:
								break
					
					if optionInstrument is None:
						raise PrepareOrderException(
							f'Could not find option instrument for {order.symbol} {leg.expiration} {leg.strike} {leg.right}', 
							order
						)
				except PrepareOrderException:
					raise
				except Exception as e:
					raise PrepareOrderException(f'Error creating option instrument for trade recovery: {e}', order)
			
			# Build the leg for the tasty trade order
			mapped_action = self._mappedOrderAction(order.action, leg.action)
			logger.debug(f'Mapping order.action={order.action} leg.action={leg.action} -> mapped_action={mapped_action}')
			comboLeg = optionInstrument.build_leg(quantity=Decimal(leg.quantity * order.quantity), action=mapped_action)
			comboLegs.append(comboLeg)
			leg.brokerSpecific['tasty_symbol'] = comboLeg.symbol  # Memorize the tastytrade symbol identifier for later transaction processing

		order.brokerSpecific['comboLegs'] = comboLegs
		order.determine_price_effect()
		return True
	
	def _transform_generic_order(self, generic_order: GenericOrder) -> NewOrder:
		"""
		Transforms the given generic order to a tastytrade order
		"""
		new_order_legs = generic_order.brokerSpecific['comboLegs']
		price = abs(generic_order.price)
		tasty_price = Decimal(price * -1 if generic_order.price_effect == PriceEffect.DEBIT else price)
		rounded_tasty_price: Decimal = round(tasty_price, 2)
		new_order = None
		if generic_order.type == OrderType.LIMIT:
			new_order = NewOrder(
				time_in_force=OrderTimeInForce.DAY,
				order_type=generic_order.type,	
				legs=new_order_legs,
				price=rounded_tasty_price
			)
		elif generic_order.type == OrderType.STOP:
			if len(generic_order.legs) == 1:
				new_order = NewOrder(
					time_in_force=OrderTimeInForce.DAY,
					order_type=generic_order.type,	
					legs=new_order_legs,
					stop_trigger=rounded_tasty_price
				)
			else:
				# Stop Orders with multiple legs must be Stop Limit Orders
				calculated_limit_price = OptionHelper.roundToTickSize(round(rounded_tasty_price * Decimal('1.25'), 2))  # Round to the nearest tick size
				tasty_limit_price: Decimal = Decimal(calculated_limit_price)
				
				new_order = NewOrder(
					time_in_force=OrderTimeInForce.DAY,
					order_type=OrderType.STOP_LIMIT,
					legs=new_order_legs,
					stop_trigger=Decimal(abs(rounded_tasty_price)),
					price=tasty_limit_price
				)
		return new_order
	
	async def placeOrder(self, managed_trade: ManagedTrade, order: GenericOrder, parent_order: GenericOrder = None) -> None:
		""" 
		Places the given order for a managed account via the broker connection.
		
		Raises:
			PlaceOrderException: If the order placement fails with a specific reason.
		"""
		account = Account.get(self._session, managed_trade.template.account)
		newOrder = self._transform_generic_order(order)
		try:
			response = account.place_order(self._session, newOrder, dry_run=False)
			#placedComplexOrders = account.get_live_complex_orders(session=self._session)
			#placedOrders = account.get_live_orders(session=self._session)
			#for order in placedOrders:
			#	logger.debug(f'Live order: {order.id} underlying: {order.underlying_symbol}')
			#	#account.delete_order(session=self._session, order_id=order.id)
			logger.debug(f'Response of place Order: {response}')
			if response.errors:
				for errorMessage in response.errors:
					raise PlaceOrderException(errorMessage, order=order)
			if response.warnings:
				for warningMessage in response.warnings:
					logger.warning(f'Warning placing order: {warningMessage}')
			
			order.brokerSpecific['order'] = response.order
			order.brokerSpecific['account'] = account
			self._orders.append(order)
			logger.debug(f'Order {response.order.id} placed successfully')
		except TastytradeError as tastyErr:
			raise PlaceOrderException(f'{tastyErr}', order=order)
		except ValidationError as valErr:
			logger.error(f'Validation error placing order: {valErr}')
			err = valErr.errors()[0]
			logger.error(err)
		except Exception as exc:
			raise PlaceOrderException(f'{exc}', order=order)

	async def place_complex_order(self, take_profit_order: GenericOrder, stop_loss_order: GenericOrder, template: Template) -> bool:
		"""
		Places the Take Profit and Stop Loss Order as complex order
		"""
		account = Account.get(self._session, template.account)
		new_order_tp = self._transform_generic_order(take_profit_order)
		new_order_sl = self._transform_generic_order(stop_loss_order)
		oco_order = NewComplexOrder( orders=[ new_order_tp, new_order_sl ] )
		try:
			dry_run = False
			response = account.place_complex_order(self._session, oco_order, dry_run=dry_run)
			logger.debug(f'Response of place_complex_order: {response}')
			if response.errors:
				for errorMessage in response.errors:
					logger.error(f'Error placing order: {errorMessage}')
					return False
			if response.warnings:
				for warningMessage in response.warnings:
					logger.warning(f'Warning placing order: {warningMessage}')
			
			try:
				take_profit_order.brokerSpecific['order'] = response.complex_order.orders[0]
				take_profit_order.brokerSpecific['account'] = account
				self._orders.append(take_profit_order)
			except IndexError as indexErr:
				logger.error(f'Error complex order. Take Profit order not found in complex order response!')
				return False

			try: 
				stop_loss_order.brokerSpecific['order'] = response.complex_order.orders[1]
				stop_loss_order.brokerSpecific['account'] = account
				self._orders.append(stop_loss_order)
			except IndexError as indexErr:
				logger.error(f'Error complex order. Stop Loss order not found in complex order response!')
				return False

			logger.debug(f'Complex Order {response.complex_order.id} placed successfully')
			return True
		except TastytradeError as tastyErr:
			logger.error(f'Error placing order: {tastyErr}')
		except ValidationError as valErr:
			logger.error(f'Validation error placing order: {valErr}')
			err = valErr.errors()[0]
			logger.error(err)
			#logger.error(repr(valErr.errors()[0]['type']))
		except Exception as exc:
			logger.error(f'Unexpected exception placing order: {exc}')
			
		return False		
		
	async def adjustOrder(self, managed_trade: ManagedTrade, order: GenericOrder, price: float) -> bool:
		""" 
		Adjusts the given order with the given new price
		"""
		if order.status == GenericOrderStatus.FILLED:
			logger.info('Order {} is already filled. Adjustment not required.', order)
			return True
		try:
			tasty_order: PlacedOrder = order.brokerSpecific['order']
		except KeyError as keyErr:
			logger.error(f'Order {order.broker_order_id} not prepared for adjustment. Cannot adjust order.')
			return False
		
		account: Account = order.brokerSpecific['account']
		logger.debug(f'Adjusting order {tasty_order.id} to price {price}')

		order.price = price
		replacement_order = self._transform_generic_order(order)

		#new_order_legs = order.brokerSpecific['comboLegs']
		#new_price = abs(price)
		#tasty_price = Decimal(new_price * -1 if order.price_effect == PriceEffect.DEBIT else new_price)

		# if order.type == OrderType.LIMIT:
		# 	replacement_order = NewOrder(
		# 		time_in_force=OrderTimeInForce.DAY,
		# 		order_type=order.type,	
		# 		legs=new_order_legs,
		# 		price=tasty_price
		# 	)
		# elif order.type == OrderType.STOP:
		# 	replacement_order = NewOrder(
		# 		time_in_force=OrderTimeInForce.DAY,
		# 		order_type=order.type,	
		# 		legs=new_order_legs,
		# 		stop_trigger=tasty_price
		# 	)
		# elif order.type == OrderType.MARKET:
		# 	replacement_order = NewOrder(
		# 		time_in_force=OrderTimeInForce.DAY,
		# 		order_type=order.type,	
		# 		legs=new_order_legs
		# 	)

		try:
			self._replacedOrders.append(tasty_order)  # Merken für das Cancel Event dieser Order
			response: PlacedOrder = account.replace_order(session=self._session, old_order_id=tasty_order.id, new_order=replacement_order)
			order.brokerSpecific['order'] = response
			#self._replacedOrders.append(response) # Auch die neue Order zu den zu ignorierenden Orders hinzufügen
			logger.debug(f'Replacment order {response.id} submitted successfully')
			return True
		
		except TastytradeError as tastyErr:
			logger.error(f'Error adjusting order: {tastyErr}')
			return False
		except ValidationError as valErr:
			logger.error(f'Validation error adjusting order: {valErr}')
			err = valErr.errors()[0]
			logger.error(err)
			return False 

	async def _perform_periodic_updates(self) -> None:
		"""
		Background job for fetching new transactions from Tastytrade for order executions.
		This is required for getting the commission and fee data.
		Additionally it checks if the access token needs to be refreshed.
		"""
		try:
			if not self.isConnected():
				return
			
			# Check if access token needs to be refreshed
			from tastytrade.utils import now_in_new_york
			if self._session.session_expiration < now_in_new_york() + timedelta(seconds=25):
				logger.debug('Access token is about to expire, refreshing...')
				self._session.refresh()

			for tasty_account in self._tasty_accounts:
				transactions =  tasty_account.get_history(self._session, start_date=dt.date.today())
				for transaction in transactions:
					if transaction.transaction_type != 'Trade' or transaction.id in self._processed_transaction_ids:
						continue
					self._processed_transaction_ids.append(transaction.id)
					await self._process_transaction(tasty_account, transaction)
		
		except Exception as e:
			logger.error(f'Error fetching new transactions: {e}')

	async def _process_transaction(self, account: Account, transaction: Transaction) -> None:
		"""
		Process a transaction from Tastytrade and report the fill and commission report events.
		"""
		logger.debug(f'Processing transaction {transaction.id}')

		# Find the corresponding generic order
		generic_order: GenericOrder = None
		tasty_order: PlacedOrder = None
		for order in self._orders:
			tasty_order = order.brokerSpecific['order']
			if tasty_order.id == transaction.order_id:
				generic_order = order
				break
		if generic_order is None:
			logger.debug(f'No matching generic order found for transaction {transaction.id} (order ID {transaction.order_id})')
			return

		# Emit a order execution details event for the transaction
		action = OrderAction.BUY if transaction.transaction_sub_type in ['Buy to Open', 'Buy to Close'] else OrderAction.SELL
		
		# Determine the matching leg of the order based on the tastytrade symbol
		matching_leg = None
		for leg in generic_order.legs:
			tasty_symbol = leg.brokerSpecific.get('tasty_symbol', None)
			if tasty_symbol == transaction.symbol:
				matching_leg = leg
				break
		if matching_leg is None:
			logger.error(f'Unexpected Error: No matching leg found in order {tasty_order.id} for transaction symbol {transaction.symbol}')
			return
		
		sec_type = 'C' if matching_leg.right == OptionRight.CALL else 'P'

		execution = Execution(
			id=str(transaction.id),
			action=action,
			sec_type=sec_type,
			strike=matching_leg.strike,
			amount=int(transaction.quantity),
			price=float(transaction.price),
			expiration=matching_leg.expiration,
			timestamp=transaction.executed_at
			)
		logger.debug(f'Emitting execution details for Tastytrade order {tasty_order.id}: {matching_leg.symbol} {matching_leg.right} {matching_leg.strike} qty={execution.amount} @ ${execution.price}')
		self._emitOrderExecutionDetailsEvent(generic_order, execution)

		# Calculate total fees and commission and emit commission report event
		commission: float = abs(float(transaction.commission))
		fee: float = float(abs(transaction.regulatory_fees) + abs(transaction.clearing_fees) + abs(transaction.proprietary_index_option_fees))
		self._emitCommissionReportEvent(order=generic_order, execution_id=str(transaction.id), commission=commission, fee=fee)

	async def _on_alert_streamer_disconnect(self, streamer: AlertStreamer):
		"""
		Callback method which is called when the Tastytrade alert streamer disconnects.
		Reconnects only the alert streamer without affecting the DXLink streamer.
		
		OTB-xxx: Added guard against multiple simultaneous reconnect attempts.
		The disconnect callback can be called multiple times during KEEPALIVE timeout
		scenarios, leading to RecursionError and multiple Alert Streamers being created.
		"""
		try:
			logger.debug("Tastytrade Alert Streamer disconnected.")
			
			# Check if we're already disconnecting the entire connection
			if self._is_disconnecting:
				logger.debug("Already disconnecting, skipping Alert Streamer reconnect")
				return
			
			# Guard against multiple simultaneous reconnect attempts
			# This is critical to prevent RecursionError and multiple streamers
			if self._is_reconnecting_alert_streamer:
				logger.debug("Alert Streamer reconnect already in progress, skipping duplicate callback")
				return
			
			# Set the reconnect flag immediately to block any other callbacks
			self._is_reconnecting_alert_streamer = True
			
			try:
				if self._streamer is None:
					logger.debug("DXLink streamer is also disconnected (None) - this is a full disconnect, not an Alert Streamer reconnect scenario")
					return
				
				logger.warning("Alert Streamer disconnected unexpectedly. Reconnecting Alert Streamer...")
				
				# CRITICAL: Close the old alert streamer FIRST to stop the async for loop
				# This prevents the old task from receiving events after being cancelled
				old_streamer = self._alert_streamer
				old_streamer_id = id(old_streamer) if old_streamer else None
				self._alert_streamer = None  # Clear reference immediately to prevent race conditions
				logger.debug(f'Set _alert_streamer to None (old streamer was {old_streamer_id})')
				
				if old_streamer:
					try:
						logger.debug(f'Closing old alert streamer {old_streamer_id}')
						await asyncio.wait_for(old_streamer.close(), timeout=2.0)
						logger.debug(f'Old alert streamer {old_streamer_id} closed successfully')
					except asyncio.TimeoutError:
						logger.warning(f'Timeout while closing old alert streamer {old_streamer_id} - continuing anyway')
					except asyncio.CancelledError:
						logger.warning(f'Close operation cancelled for alert streamer {old_streamer_id} - continuing anyway')
					except Exception as exc:
						logger.warning(f'Error closing old alert streamer {old_streamer_id}: {exc} - continuing anyway')
				
				# Now cancel the account listening task (it should exit cleanly now)
				old_task = self.task_listen_accounts
				old_task_id = id(old_task) if old_task else None
				current_task = asyncio.current_task()
				current_task_id = id(current_task) if current_task else None
				
				# Clear the reference before cancelling to prevent self-referencing issues
				self.task_listen_accounts = None
				
				# Check if we're trying to cancel ourselves (which would cause recursion)
				if old_task and old_task is current_task:
					logger.debug(f'Task {old_task_id} is the current task, skipping self-cancel')
				elif old_task and not old_task.done():
					logger.debug(f'Cancelling account listening task {old_task_id} (current task is {current_task_id})')
					old_task.cancel()
					# Wait for the task using shield and gather to avoid propagating CancelledError
					try:
						# Use return_exceptions=True to prevent exceptions from being raised
						await asyncio.wait_for(
							asyncio.gather(old_task, return_exceptions=True),
							timeout=2.0
						)
						logger.debug(f'Account listening task {old_task_id} cancelled successfully')
					except asyncio.TimeoutError:
						logger.warning(f'Timeout waiting for account listening task {old_task_id} to cancel - continuing anyway')
					except Exception as exc:
						logger.warning(f'Error waiting for account listening task to cancel: {exc} - continuing anyway')
				else:
					logger.debug(f'Account listening task already done or None, no need to cancel')
				
				# Small delay to ensure everything is cleaned up
				logger.debug('Waiting 0.5s for cleanup before reconnecting')
				await asyncio.sleep(0.5)
				
				# Restart the account updates task which will create a new alert streamer
				try:
					# Check if event loop is still running before creating new task
					loop = asyncio.get_event_loop()
					if loop.is_running():
						self.task_listen_accounts = asyncio.create_task(self._request_account_updates())
						new_task_id = id(self.task_listen_accounts)
						logger.info(f"Alert Streamer reconnected successfully (new task {new_task_id})")
					else:
						logger.debug("Event loop not running, skipping Alert Streamer reconnect")
				except RuntimeError as runtime_err:
					logger.warning(f'Cannot reconnect Alert Streamer: {runtime_err}')
				except Exception as exc:
					logger.error(f'Error reconnecting Alert Streamer: {exc}')
			finally:
				# Always reset the reconnect flag when done
				self._is_reconnecting_alert_streamer = False
		except Exception as e:
			self._is_reconnecting_alert_streamer = False
			logger.error(f"Critical error in _on_alert_streamer_disconnect: {e}", exc_info=True)

	async def _handle_alert_streamer_error(self, error: Exception):
		"""
		Handle fatal errors from the AlertStreamer (e.g., AUTH timeout).
		
		OTB-264: When the AlertStreamer encounters a fatal error that doesn't trigger
		the disconnect callback, we need to manually initiate a reconnect.
		
		This handles cases like:
		- "The timeout for waiting for AUTH has been reached"
		- Connection errors during streaming
		"""
		try:
			logger.warning(f'AlertStreamer fatal error: {error}. Attempting reconnect...')
			
			# Check if we should attempt reconnect
			if self._is_disconnecting:
				logger.debug('Already disconnecting, skipping AlertStreamer reconnect after error')
				return
			
			# Guard against multiple simultaneous reconnect attempts
			if self._is_reconnecting_alert_streamer:
				logger.debug('Alert Streamer reconnect already in progress, skipping error-triggered reconnect')
				return
			
			self._is_reconnecting_alert_streamer = True
			
			try:
				if self._streamer is None:
					logger.debug('DXLink streamer is None - this may be a full disconnect scenario')
					# Emit disconnected event to trigger full reconnection
					self._emitDisconnectedEvent()
					return
				
				# Clean up the old alert streamer
				old_streamer = self._alert_streamer
				self._alert_streamer = None
				
				if old_streamer:
					try:
						await asyncio.wait_for(old_streamer.close(), timeout=2.0)
					except Exception as close_err:
						logger.debug(f'Error closing old alert streamer after error: {close_err}')
				
				# Wait a bit before reconnecting to avoid rapid reconnect loops
				await asyncio.sleep(2.0)
				
				# Check again if we should reconnect (state may have changed during sleep)
				if self._is_disconnecting or self._streamer is None:
					logger.debug('State changed during wait, skipping AlertStreamer reconnect')
					return
				
				# Create a new task to handle account updates (which creates a new AlertStreamer)
				try:
					loop = asyncio.get_event_loop()
					if loop.is_running():
						self.task_listen_accounts = asyncio.create_task(self._request_account_updates())
						logger.info('AlertStreamer reconnect initiated after fatal error')
					else:
						logger.debug('Event loop not running, cannot reconnect AlertStreamer')
				except RuntimeError as runtime_err:
					logger.warning(f'Cannot reconnect AlertStreamer: {runtime_err}')
			finally:
				self._is_reconnecting_alert_streamer = False
				
		except Exception as e:
			self._is_reconnecting_alert_streamer = False
			logger.error(f'Error handling AlertStreamer error: {e}', exc_info=True)

	async def _on_streamer_disconnect(self, streamer: DXLinkStreamer):
		"""
		Callback method which is called when the Tastytrade streamer disconnects.
		
		OTB-236: This callback may be invoked by the tastytrade library during shutdown
		when the event loop is already closing. We need to:
		1. Check if event loop is still running
		2. Avoid calling _disconnect_internal() which would try to close the streamer again (recursion!)
		3. Handle unexpected disconnects without causing infinite loops
		"""
		try:
			logger.debug(f'Tastytrade Streamer disconnect callback triggered (_is_disconnecting={self._is_disconnecting})')
			
			if self._backgroundScheduler.running:
				self._backgroundScheduler.remove_all_jobs()
				self._backgroundScheduler.shutdown(wait=False)

			# Check if we're already in the process of disconnecting
			if self._is_disconnecting:
				logger.debug('Already disconnecting, ignoring streamer disconnect callback')
				return
			
			# OTB-236: Check if event loop is still running to avoid RuntimeError during shutdown
			try:
				loop = asyncio.get_running_loop()  # noqa: F841
				# Note: is_closing() is not available in Python 3.11, only check if loop exists
				logger.debug('Event loop is running')
			except RuntimeError as e:
				logger.debug(f'Event loop not running, ignoring streamer disconnect callback (shutdown in progress): {e}')
				return
			
			# Handle unexpected disconnect (not during our controlled shutdown)
			logger.warning('Unexpected DXLink streamer disconnect detected. Initiating full reconnection...')
			logger.debug(f'Streamer: {id(self._streamer) if self._streamer else None}, Alert Streamer: {id(self._alert_streamer) if self._alert_streamer else None}')
			
			# OTB-236: Set flag BEFORE any operations to prevent recursive calls
			self._is_disconnecting = True
			
			# OTB-236: Don't call _disconnect_internal() here as it will try to close the streamer again!
			# Instead, manually clean up and trigger reconnection via the disconnected event
			
			# Disable trading immediately
			await self.set_trading_enabled(False, 'Streamer disconnected unexpectedly')
			
			# Cancel streaming tasks - use gather with return_exceptions to avoid recursion
			tasks_to_cancel = [
				task for task in [
					self.task_listen_quotes,
					self.task_listen_greeks,
					self.task_listen_candle,
					self.task_listen_accounts
				] if task and not task.done()
			]
			
			if tasks_to_cancel:
				# Cancel all tasks
				for task in tasks_to_cancel:
					task.cancel()
				
				# Wait for cancellation to complete
				try:
					await asyncio.wait_for(
						asyncio.gather(*tasks_to_cancel, return_exceptions=True),
						timeout=2.0
					)
					logger.debug(f'All {len(tasks_to_cancel)} streaming tasks cancelled')
				except asyncio.TimeoutError:
					logger.debug(f'Timeout waiting for streaming tasks to cancel')
				except Exception as e:
					logger.debug(f'Error waiting for streaming tasks to cancel: {e}')
			
			# Clear task references
			self.task_listen_quotes = None
			self.task_listen_greeks = None
			self.task_listen_candle = None
			self.task_listen_accounts = None
			
			# Clear streamer references (don't call close() - it's already closing!)
			self._streamer = None
			self._alert_streamer = None
			
			# Emit the disconnected event to trigger reconnection logic
			self._emitDisconnectedEvent()
		except Exception as e:
			logger.error(f'Unexpected error in _on_streamer_disconnect: {e}', exc_info=True)
	
	async def _subscribe_data(self) -> None:
		"""
		Subscribe to the required data
		"""
		logger.debug('Subscribing to streaming data')
		await self._streamer.subscribe(Quote, self._quote_symbols)
		await self._streamer.subscribe(Greeks, self._greeks_symbols)
		start_time = dt.datetime.now() - timedelta(days=1)
		await self._streamer.subscribe_candle(self._candle_symbols, interval='1m', start_time=start_time)

	async def requestTickerData(self, symbols: List[str]) -> None:
		""" 
		Request ticker data for the given symbols and their options
		"""
		# Check if this is the first call (initial setup) or a reconnect/re-subscription
		# OTB-xxx: Use _symbolData to distinguish between initial setup and reconnect
		# After a reconnect, _streamer is None but _symbolData still contains the previous data
		is_initial_setup = self._streamer is None and len(self._symbolData) == 0
		is_reconnect = self._streamer is None and len(self._symbolData) > 0
		
		if is_initial_setup or is_reconnect:
			ssl_context = ssl.create_default_context(cafile=certifi.where())
			self._streamer = await DXLinkStreamer(self._session, ssl_context=ssl_context, disconnect_fn=self._on_streamer_disconnect)

		if is_initial_setup:
			# Initial setup: Create new symbolData objects
			self._quote_symbols = []
			self._candle_symbols = []
			self._greeks_symbols = symbols

			for symbol in symbols:
				match symbol:
					case 'SPX':
						symbolData = TastySymbolData()
						symbolData.symbol = symbol
						symbolData.tastySymbol = 'SPX'
						self._quote_symbols.append('SPX')
						self._symbolData[symbol] = symbolData
						self._symbolReverseLookup[symbolData.tastySymbol] = symbol
					case 'VIX':
						symbolData = TastySymbolData()
						symbolData.symbol = symbol
						symbolData.tastySymbol = 'VIX'
						symbolData.trade_options = False  # No options for VIX
						self._candle_symbols.append('VIX')
						self._symbolData[symbol] = symbolData
						self._symbolReverseLookup[symbolData.tastySymbol] = symbol
					case _:
						logger.error(f'Symbol {symbol} currently not supported by Tastytrade Connector!')
						continue

			# Request option chains and expirations for symbols with options trading enable
			current_date = dt.date.today()
			for item in self._symbolData.values():
				symbol_data: TastySymbolData = item
				desired_chain = None
				if symbol_data.trade_options:
					# If options are enabled, request the option chain and expirations
					symbol_information = symbolInfo.symbol_infos[symbol_data.symbol]
					chains = NestedOptionChain.get(self._session, symbol_data.tastySymbol)
					for chain in chains:
						if chain.root_symbol == symbol_information.trading_class:
							desired_chain = chain
							break
					if not desired_chain:
						logger.error(f'No option chain found for symbol {symbol_data.symbol} with trading class {symbol_information.trading_class}!')
						continue
					symbol_data.chain = desired_chain

					got_strikes = False
					for expiration in desired_chain.expirations:
						if expiration.expiration_date < current_date:
							continue
						symbol_data.expirations.append(expiration.expiration_date)
						if not got_strikes:
							# Obtain strikes of the first expiration only
							for strike in expiration.strikes:
								symbol_data.strikes.append(float(strike.strike_price))
							got_strikes = True

					if current_date > symbol_data.expirations[0]:
						logger.warning(f'There are no {symbol_data.symbol} options expiring today!')

			# OTB-253: Symbol data is now ready for trade recovery to use
			# Start subscribe operations in background (don't await - they can take 2+ seconds)
			# This prevents blocking trade recovery which needs symbol data but not live streaming yet
			logger.debug('Symbol data loaded, starting streaming subscriptions in background')
			
			# Start subscribe and streaming tasks in background
			asyncio.create_task(self._start_streaming_async())
		elif is_reconnect:
			# OTB-xxx: Reconnect after disconnect - keep existing symbolData with optionPriceData
			# but re-create the streamer and re-subscribe to data
			logger.debug('Reconnecting to ticker data (keeping existing symbolData with price data)')
			
			# Start subscribe and streaming tasks in background
			asyncio.create_task(self._start_streaming_async())
		else:
			# Re-subscription after EOD settlement: Keep existing symbolData, just re-subscribe
			logger.debug('Re-subscribing to ticker data after EOD settlement (keeping existing symbolData)')
			
			# Re-subscribe to base symbols (SPX, VIX) - these should still be in the lists
			await self._subscribe_data()
			
			# Mark ticker data as subscribed again
			self._subscribed_ticker_data = True
			logger.debug('Ticker data re-subscription completed')

	async def _start_streaming_async(self) -> None:
		"""
		OTB-253: Start streaming subscriptions and tasks asynchronously.
		This method runs in the background to avoid blocking trade recovery.
		"""
		try:
			# Subscribe to data (this can take 2+ seconds due to network calls)
			await self._subscribe_data()
			self._subscribed_ticker_data = True
			
			# OTB-xxx: Start streaming tasks but don't wait for them (they run indefinitely)
			# These tasks will be cancelled during disconnect via _disconnect_internal()
			if self.task_listen_quotes is None:
				self.task_listen_quotes = asyncio.create_task(self._update_quotes())
				self.task_listen_greeks = asyncio.create_task(self._update_greeks())
				self.task_listen_candle = asyncio.create_task(self._update_candle())
				logger.debug('Started streaming tasks for quotes, greeks, and candles')
		except Exception as e:
			logger.error(f'Failed to start streaming: {e}', exc_info=True)

	async def _update_accounts(self):
		"""
		Task for listening to account updates
		"""
		logger.debug(f'_update_accounts started, alert_streamer={id(self._alert_streamer) if self._alert_streamer else None}')
		try:
			# Check if streamer is still valid before entering the loop
			if self._alert_streamer is None:
				logger.debug('Alert streamer is None, cannot start account updates loop')
				return
			
			streamer_id = id(self._alert_streamer)
			logger.debug(f'Starting async for loop with streamer {streamer_id}')
			async for order in self._alert_streamer.listen(PlacedOrder):
				logger.debug(f'Received order event in loop (streamer {streamer_id}): {order.id} status {order.status}')
				additional_info = ''
				if order.status == OrderStatus.REJECTED:
					additional_info = f'Reason: {order.reject_reason}'
				logger.debug(f'Update on order {order.id} status {order.status} {additional_info}')			
				ignore_order_event = False
				# Cancel Events von Preisanpassungen ignorieren, da sie kein echtes Cancel sind
				for replaced_order in self._replacedOrders:
					if replaced_order.id == order.id and order.status == OrderStatus.CANCELLED:
						#self._replacedOrders.remove(replaced_order)
						ignore_order_event = True
						logger.debug('Ignoring cancel event for replaced order')
						continue
					if replaced_order.id == order.id and (order.status == OrderStatus.ROUTED or order.status == OrderStatus.LIVE):
						if order.status == OrderStatus.LIVE:
							self._replacedOrders.remove(replaced_order)
						ignore_order_event = True
						logger.debug('Ignoring placement of new replacement order')
						continue
		
				if not ignore_order_event:
					relevantOrder: GenericOrder = None
					for managedOrder in self._orders:
						broker_specific_order: PlacedOrder = managedOrder.brokerSpecific['order']
						if broker_specific_order.id == order.id:
							relevantOrder = managedOrder
							break
				
					if relevantOrder == None:
						logger.debug(f'No managed order matched the status event')
					else:
						relevantOrder.brokerSpecific['order'] = order
						filledAmount = int(order.size)
						relevantOrder.filledQuantity = filledAmount  # Track filled quantity for partial fill handling
						relevantOrder.averageFillPrice = abs(float(order.price)) if order.price != None else 0
						order_status =	self._genericOrderStatus(order.status)
						if order_status and order_status != relevantOrder.status:
							self._emitOrderStatusEvent(relevantOrder, order_status, filledAmount)
							# Emit execution details when order is filled
							#if order.status == OrderStatus.FILLED:
							#	logger.debug(f'Order {order.id} is filled, emitting execution details')
							#	self._emit_execution_details_for_order(relevantOrder, order)
						else:
							logger.debug(f'Order status {order.status} not mapped to generic order status')
		except asyncio.CancelledError:
			logger.debug(f'Account updates task cancelled, stopping order listener (streamer {id(self._alert_streamer) if self._alert_streamer else None})')
			raise  # Re-raise to allow proper task cancellation
		except Exception as exc:
			logger.error(f'Unexpected error in _update_accounts: {exc}', exc_info=True)
		finally:
			logger.debug(f'_update_accounts exiting (streamer {id(self._alert_streamer) if self._alert_streamer else None})')

	async def _update_quotes(self):
		async for e in self._streamer.listen(Quote):
			logger.trace(f'Received Quote: {e.event_symbol} bid price: {e.bid_price} ask price: {e.ask_price}')
			# Preisdaten speichern if the broker is subscribed to ticker data
			if self._subscribed_ticker_data is None:
				continue

			if not e.event_symbol.startswith('.'):
				# Symbol ist ein Basiswert
				try:
					genericSymbol = self._symbolReverseLookup[e.event_symbol]
					symbolData: TastySymbolData  = self._symbolData[genericSymbol]
					midPrice = float((e.bid_price + e.ask_price) / 2)
					atmStrike = OptionHelper.roundToStrikePrice(midPrice)
					symbolData.lastPrice = midPrice
					if symbolData.lastAtmStrike != atmStrike:  # Check for missing Option Data only if ATM Strike has changed
						symbolData.lastAtmStrike = atmStrike
						asyncio.create_task(self._requestMissingOptionData(symbolData, atmStrike))
					
				except KeyError as keyErr:
					logger.error(f'No generic symbol found for tastytrade symbol {e.event_symbol}')
			else:
				# Symbol ist eine Option
				try:
					symbol, optionType, expiration, strike = self._getOptionInfos(e.event_symbol)
					symbol_information = symbolInfo.symbol_infos[symbol]
					symbolData = self._symbolData[symbol]
					optionStrikeData = symbolData.optionPriceData[expiration]
					optionStrikePriceData = optionStrikeData.strikeData[strike]
					if optionType == OptionType.CALL:
						optionStrikePriceData.callBid = float(e.bid_price)
						optionStrikePriceData.callAsk = float(e.ask_price)
					else:
						optionStrikePriceData.putBid = float(e.bid_price)
						optionStrikePriceData.putAsk = float(e.ask_price)
					optionStrikePriceData.lastUpdated = dt.datetime.now(symbol_information.timezone)
					self._last_option_price_update_time = optionStrikePriceData.lastUpdated
					await self._monitor_bid_ask_spread(optionStrikePriceData, genericSymbol, strike)
				except Exception as exc:
					logger.error(f'Error getting option infos: {exc}')
	
	async def _update_greeks(self):
		async for e in self._streamer.listen(Greeks):
			logger.trace(f'Received Greeks: {e.event_symbol} delta: {e.delta}')
			if e.event_symbol.startswith('.'):
				# Symbol ist eine Option
				try:
					symbol, optionType, expiration, strike = self._getOptionInfos(e.event_symbol)
					symbol_information = symbolInfo.symbol_infos[symbol]
					symbolData = self._symbolData[symbol]
					optionStrikeData = symbolData.optionPriceData[expiration]
					optionStrikePriceData = optionStrikeData.strikeData[strike]
					if optionType == OptionType.CALL:
						optionStrikePriceData.callDelta = float(e.delta)
					else:
						optionStrikePriceData.putDelta = float(e.delta)

				except Exception as exc:
					logger.error(f'Error getting option infos: {exc}')

	async def _update_candle(self):
		async for e in self._streamer.listen(Candle):
			logger.trace(f'Received Candle: {e.event_symbol} close: {e.close}')
			if not e.event_symbol.startswith('.'):
				# Symbol ist ein Basiswert
				try:
					symbol = e.event_symbol.split('{')[0]
					genericSymbol = self._symbolReverseLookup[symbol]
					symbolData: TastySymbolData  = self._symbolData[genericSymbol]
					symbolData.lastPrice = float(e.close)
				except KeyError as keyErr:
					logger.error(f'No generic symbol found for tastytrade symbol {e.event_symbol}')

	async def eod_settlement_tasks(self) -> None:
		"""
		Perform End of Day settlement tasks
		"""
		await super().eod_settlement_tasks()

	def get_option_strike_data(self, symbol: str, expiration: dt.date) -> OptionStrikeData:
		""" 
		Returns the option strike data for the given symbol and expiration. It is including
		prices and greeks.
		"""
		symbolData = self._symbolData[symbol]
		try:
			return symbolData.optionPriceData[expiration]
		except KeyError:
			raise ValueError(f'No option strike data for symbol {symbol} and expiration {expiration} found!')

	def get_option_strike_price_data(self, symbol: str, expiration: date, strike: float) -> OptionStrikePriceData:
		""" 
		Returns the option strike price data for the given symbol, expiration date, strike price and right.
		Returns None if no price data is available for the expiration date or strike.
		"""
		symbolData = self._symbolData[symbol]
		try:
			optionStrikeData = symbolData.optionPriceData[expiration]
		except KeyError:
			# No price data available for this expiration date (e.g., after trade recovery)
			return None
		
		if strike in optionStrikeData.strikeData.keys():
			return optionStrikeData.strikeData[strike]
		else:
			return None

	def get_strike_by_delta(self, symbol: str, right: str, delta: int) -> float:
		"""
		Returns the strike price based on the given delta based on the buffered option price data
		"""
		symbolData = self._symbolData[symbol]
		current_date = dt.date.today()
		option_price_data: OptionStrikeData= symbolData.optionPriceData[current_date]
		previous_delta = 0
		previous_strike = 0
		reverse = True if right == OptionRight.PUT else False
		sorted_strikes = dict(sorted(option_price_data.strikeData.items(), reverse=reverse))
		for strike, price_data in sorted_strikes.items():
			if right == OptionRight.PUT:
				if price_data.putDelta == None:
					continue
				adjusted_delta = price_data.putDelta * -100
			else:
				if price_data.callDelta == None:
					continue
				adjusted_delta = price_data.callDelta * 100
			if adjusted_delta <= delta:
				if OptionHelper.closest_number(delta, previous_delta, adjusted_delta) == adjusted_delta:
					return strike
				else:
					return previous_strike
			previous_delta = adjusted_delta
			previous_strike = strike

		raise ValueError(f'No strike price found for delta {delta} in symbol {symbol}!')
	
	def get_strike_by_price(self, symbol: str, right: str, price: float) -> float:
		""" 
		Returns the strike price based on the given premium price based on the buffered option price data
		"""
		# TODO: Implement in base class, because TWS Connector got the same code
		symbolData = self._symbolData[symbol]
		current_date = dt.date.today()
		option_price_data: OptionStrikeData= symbolData.optionPriceData[current_date]
		previous_price = 0
		previous_strike = 0
		reverse = True if right == OptionRight.PUT else False
		sorted_strikes = dict(sorted(option_price_data.strikeData.items(), reverse=reverse))
		for strike, price_data in sorted_strikes.items():
			if right == OptionRight.PUT:
				current_strike_price = price_data.getPutMidPrice()
				if current_strike_price == None:
					continue
			elif right == OptionRight.CALL:
				current_strike_price = price_data.getCallMidPrice()
				if current_strike_price == None:
					continue
			if current_strike_price > 0 and current_strike_price <= price:
				if OptionHelper.closest_number(price, previous_price, current_strike_price) == current_strike_price:
					return strike
				else:
					return previous_strike
			previous_price = current_strike_price
			previous_strike = strike
		raise ValueError(f'No strike price found for price {price} in symbol {symbol}!')

	def getFillPrice(self, order: GenericOrder) -> float:
		""" 
		Returns the fill price of the given order if it is filled
		"""
		try:
			tastyOrder: PlacedOrder = order.brokerSpecific['order']
			if tastyOrder.status == OrderStatus.FILLED:
				return abs(float(tastyOrder.price))
			else:
				return 0
		except KeyError as keyErr:
			logger.error(f'No fill price available for order {order}')
	
	def getLastPrice(self, symbol: str) -> float:
		""" 
		Returns the last price of the given symbol
		"""
		try:
			symbolData = self._symbolData[symbol]
			return symbolData.lastPrice
		except KeyError as keyErr:
			logger.error(f'No last price available for symbol {symbol}')
			return 0

	def oco_as_complex_order(self) -> bool:
		"""
		With Tastytrade, the OCO orders have to be placed as one complex order
		"""
		return True
	
	def uses_oco_orders(self) -> bool:
		""" 
		The TWS Connector uses OCO orders for take profit and stop loss orders
		"""
		return True
	
	async def unsubscribe_ticker_data(self):
		"""
		Unsubscribe from ticker data - including base symbols and all options
		"""
		await super().unsubscribe_ticker_data()
		
		# First, unsubscribe from all option symbols
		for symbolData in self._symbolData.values():
			try:
				for day_option_price_data in symbolData.optionPriceData.values():
					streamer_symbols = []
					for value in day_option_price_data.strikeData.values():
						option_price_data: OptionStrikePriceData = value
						
						try:
							call_streamer_symbol = option_price_data.brokerSpecific['call_streamer_symbol']
							streamer_symbols.append(call_streamer_symbol)
							self._quote_symbols.remove(call_streamer_symbol)
						except KeyError:
							pass

						try:
							put_streamer_symbol = option_price_data.brokerSpecific['put_streamer_symbol']
							streamer_symbols.append(put_streamer_symbol)
							self._quote_symbols.remove(put_streamer_symbol)
						except KeyError:
							pass

					if streamer_symbols:
						try:
							await self._streamer.unsubscribe(Quote, streamer_symbols)
							await self._streamer.unsubscribe(Greeks, streamer_symbols)
						except ConnectionClosedOK:
							logger.debug('Streamer already closed, ignore')
							pass
				symbolData.optionPriceData.clear()
				symbolData.lastAtmStrike = 0
				symbolData.lastPrice = 0
			except KeyError:
				pass
		
		# Now unsubscribe from base symbols (SPX, VIX, etc.)
		# This ensures a clean state for re-subscription at new day start
		if self._quote_symbols:
			logger.debug(f'Unsubscribing from {len(self._quote_symbols)} base quote symbols')
			await self._streamer.unsubscribe(Quote, self._quote_symbols)
		if self._greeks_symbols:
			logger.debug(f'Unsubscribing from {len(self._greeks_symbols)} base greeks symbols')
			await self._streamer.unsubscribe(Greeks, self._greeks_symbols)
		if self._candle_symbols:
			logger.debug(f'Unsubscribing from {len(self._candle_symbols)} base candle symbols')
			await self._streamer.unsubscribe(Candle, self._candle_symbols)

	async def _requestMissingOptionData(self, symbolData: TastySymbolData, atmStrike: float):
		"""
		Request option data for the given symbol and expiration date
		"""
		relevant_expirations = self._determine_relevant_expirations(symbolData)
		symbolInformation = symbolInfo.symbol_infos[symbolData.symbol]
		strikes_of_interest = self._determine_strikes_of_interest(symbolData, atmStrike)	
	
		for relevant_expiration in relevant_expirations:
			# Optain saved options data for the expiration date
			try:
				#expiration_date_str = relevant_expiration.strftime('%Y%m%d')
				optionStrikeData = symbolData.optionPriceData[relevant_expiration]
			except KeyError as keyErr:
				# Wenn noch keine Optionsdaten für das Verfallsdatum vorhanden sind, dann bei Tasty anfragen ob es Optionsdaten gibt

				optionStrikeData = OptionStrikeData()
				symbolData.optionPriceData[relevant_expiration] = optionStrikeData

			for chain_at_expiration in symbolData.chain.expirations:
				if chain_at_expiration.expiration_date >= relevant_expiration:
					break
			if chain_at_expiration == None or chain_at_expiration.expiration_date != relevant_expiration:
				logger.error(f'No options available for symbol {symbolData.tastySymbol} and expiration date {relevant_expiration}')
				continue
			
			# Convert the available strikes to a list of decimal numbers
			# if not done yet
			try:
				available_strikes = chain_at_expiration._optrabot_strikes
			except AttributeError as attrErr:
				chain_at_expiration._optrabot_strikes = []
				available_strikes = chain_at_expiration._optrabot_strikes
				for strike in chain_at_expiration.strikes:
					available_strikes.append(strike.strike_price)

			options_to_be_requested = []
			for strike_price in strikes_of_interest:	
				try:
					optionStrikeData.strikeData[strike_price]
				except KeyError as keyErr:
					option_strike_price_data = OptionStrikePriceData()
					optionStrikeData.strikeData[strike_price] = option_strike_price_data
					options_to_be_requested.append(strike_price)

			if len(options_to_be_requested) > 0:
				streamer_symbols = []
				for item in chain_at_expiration.strikes:
					strike: Strike = item
					if strike.strike_price in options_to_be_requested:
						option_strike_data = optionStrikeData.strikeData[strike.strike_price]
						option_strike_data.brokerSpecific['call_option'] = strike.call
						option_strike_data.brokerSpecific['call_streamer_symbol'] = strike.call_streamer_symbol
						option_strike_data.brokerSpecific['put_option'] =  strike.put
						option_strike_data.brokerSpecific['put_streamer_symbol']  = strike.put_streamer_symbol
						streamer_symbols.append(strike.call_streamer_symbol)
						streamer_symbols.append(strike.put_streamer_symbol)
						self._quote_symbols.append(strike.call_streamer_symbol)
						self._quote_symbols.append(strike.put_streamer_symbol)
				await self._streamer.subscribe(Quote, streamer_symbols)
				await self._streamer.subscribe(Greeks, streamer_symbols)
		
	def _getOptionInfos(self, tastySymbol: str) -> tuple:
		"""
		Extracts the generic symbol and expiration date, strike and option side from the tastytrade option symbol.
		If the option symbol information cannot be parsed as expected, a ValueError exception is raised.
		"""
		error = False
		pattern = r'^.(?P<optionsymbol>[A-Z]+)(?P<expiration>[0-9]+)(?P<type>[CP])(?P<strike>[0-9]+)'
		compiledPattern = re.compile(pattern)
		match = compiledPattern.match(tastySymbol)
		try:
			if match:
				optionSymbol = match.group('optionsymbol')
				for symbol, symbol_info in symbolInfo.symbol_infos.items():
					if symbol_info.symbol + symbol_info.option_symbol_suffix == optionSymbol:
						genericSymbol = symbol_info.symbol
						break
				expirationDate = dt.datetime.strptime(match.group('expiration'), '%y%m%d').date()
				strike = float(match.group('strike'))
				optionType = OptionType.CALL if match.group('type') == 'C' else OptionType.PUT
		except IndexError as indexErr:
			logger.error(f'Invalid option symbol {tastySymbol}')
			error = True
		except ValueError as valueErr:
			logger.error(f'Invalid option symbol {tastySymbol}')
			error = True
		if genericSymbol == None or error == True:
			raise ValueError(f'Invalid option symbol {tastySymbol}')
		return genericSymbol, optionType, expirationDate, strike
	
	def _mappedOrderAction(self, orderAction: GenericOrderAction, legAction: GenericOrderAction) -> OrderAction:
		"""
		Maps the general order action to the Tasty specific order action
		For Closing orders, the leg actions need to be inverted
		eg. BUY -> SELL_TO_CLOSE
		eg. SELL -> BUY_TO_CLOSE
		"""
		match orderAction:
			case GenericOrderAction.BUY_TO_OPEN:
				if legAction == GenericOrderAction.BUY:
					return OrderAction.BUY_TO_OPEN
				if legAction == GenericOrderAction.SELL:
					return OrderAction.SELL_TO_OPEN
				else:
					raise ValueError(f'Unknown leg action: {legAction}')
			case GenericOrderAction.SELL_TO_OPEN:
				if legAction == GenericOrderAction.SELL:
					return OrderAction.SELL_TO_OPEN
				elif legAction == GenericOrderAction.BUY:
					return OrderAction.BUY_TO_OPEN
				else:
					raise ValueError(f'Unknown leg action: {legAction}')
			case GenericOrderAction.BUY_TO_CLOSE:
				if legAction == GenericOrderAction.BUY:
					return OrderAction.BUY_TO_CLOSE
				if legAction == GenericOrderAction.SELL:
					return OrderAction.SELL_TO_CLOSE
				else:
					raise ValueError(f'Unknown leg action: {legAction}')
			case GenericOrderAction.SELL_TO_CLOSE:
				if legAction == GenericOrderAction.SELL:
					return OrderAction.SELL_TO_CLOSE
				elif legAction == GenericOrderAction.BUY:
					return OrderAction.BUY_TO_CLOSE
				else:
					raise ValueError(f'Unknown leg action for SELL_TO_CLOSE: {legAction}')
			case _:
				raise ValueError(f'Unknown order action: {orderAction}')
			
	async def _request_account_updates(self):
		"""
		Request Account Updates - runs as a background task.
		
		If the AlertStreamer encounters a fatal error (e.g., AUTH timeout),
		this method will attempt to reconnect automatically.
		"""
		task_id = id(asyncio.current_task())
		logger.debug(f'_request_account_updates started (task {task_id})')
		try:
			self._alert_streamer = await AlertStreamer(self._session, disconnect_fn=self._on_alert_streamer_disconnect)
			streamer_id = id(self._alert_streamer)
			logger.debug(f'Created alert streamer {streamer_id} (task {task_id})')
			await self._alert_streamer.subscribe_accounts(self._tasty_accounts)
			
			# Start the update loop directly without creating nested tasks
			logger.debug(f'Starting _update_accounts (task {task_id}, streamer {streamer_id})')
			await self._update_accounts()
		except asyncio.CancelledError:
			logger.debug(f'Cancelled listening to account updates (task {task_id})')
			raise  # Re-raise to allow proper task cancellation
		except Exception as exc:
			logger.error(f'Error in account updates (task {task_id}): {exc}')
			# OTB-264: Attempt to reconnect the AlertStreamer on fatal errors
			# (e.g., "The timeout for waiting for AUTH has been reached")
			await self._handle_alert_streamer_error(exc)
		finally:
			logger.debug(f'_request_account_updates exiting (task {task_id})')

	def _genericOrderStatus(self, status: OrderStatus) -> GenericOrderStatus:
		"""
		Maps the Tastytrade order status to the generic order status
		"""
		match status:
			case OrderStatus.RECEIVED:
				return GenericOrderStatus.OPEN
			case OrderStatus.LIVE:
				return GenericOrderStatus.OPEN
			case OrderStatus.CONTINGENT:
				return GenericOrderStatus.OPEN
			case OrderStatus.CANCELLED:
				return GenericOrderStatus.CANCELLED
			case OrderStatus.FILLED:
				return GenericOrderStatus.FILLED
			case OrderStatus.REJECTED:
				return GenericOrderStatus.CANCELLED
			case _:
				return None


"""Add settings table for key-value storage

Revision ID: fca33d484976
Revises: a1b2c3d4e5f6
Create Date: 2025-10-24 08:23:56.166732

"""
from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = 'fca33d484976'
down_revision: Union[str, None] = 'a1b2c3d4e5f6'
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    """Create settings table for key-value storage"""
    op.create_table(
        'settings',
        sa.Column('key', sa.String(), nullable=False),
        sa.Column('value', sa.String(), nullable=True),
        sa.Column('description', sa.String(), nullable=True),
        sa.Column('updated_at', sa.TIMESTAMP(), nullable=True),
        sa.PrimaryKeyConstraint('key')
    )


def downgrade() -> None:
    """Drop settings table"""
    op.drop_table('settings')




"""Spread Monitoring Configuration

This module provides configuration types for bid/ask spread monitoring.
Introduced with OTB-284 enhancement for configurable spread monitoring.

The spread monitoring feature detects when the bid/ask spread of an option
exceeds a configurable threshold, which may indicate illiquid market conditions
or potential pricing issues.

Example config.yaml:
    general:
      spread_monitoring:
        mode: all  # 'off', 'active_strikes', or 'all'
        multiplier: 2.0
        min_ask_threshold: 0.30
"""

from dataclasses import dataclass
from enum import Enum


class SpreadMonitoringMode(Enum):
    """
    Defines the monitoring mode for bid/ask spread checks.
    
    Attributes:
        OFF: Spread monitoring is completely disabled
        ACTIVE_STRIKES: Only monitor strikes that are part of active trades
        ALL: Monitor all strikes in the option chain (default behavior)
    """
    OFF = 'off'
    ACTIVE_STRIKES = 'active_strikes'
    ALL = 'all'


@dataclass
class SpreadMonitoringConfig:
    """
    Configuration for bid/ask spread monitoring.
    
    The spread is considered "high" when:
        spread = ask_price - bid_price
        spread > bid_price * multiplier
    
    Attributes:
        mode: Determines which strikes to monitor (default: OFF)
        multiplier: The ask/bid ratio threshold for high spread detection
        min_ask_threshold: Minimum ask price to trigger monitoring
    """
    mode: SpreadMonitoringMode = SpreadMonitoringMode.OFF
    multiplier: float = 2.0
    min_ask_threshold: float = 0.30
    
    @classmethod
    def from_config_dict(cls, config_dict: dict | None) -> 'SpreadMonitoringConfig':
        """
        Creates a SpreadMonitoringConfig from a configuration dictionary.
        
        Args:
            config_dict: Dictionary with spread_monitoring configuration
            
        Returns:
            SpreadMonitoringConfig instance with values from dict or defaults
            
        Example:
            >>> config_dict = {'mode': 'active_strikes', 'multiplier': 1.5}
            >>> config = SpreadMonitoringConfig.from_config_dict(config_dict)
            >>> config.mode
            <SpreadMonitoringMode.ACTIVE_STRIKES: 'active_strikes'>
        """
        if not config_dict:
            return cls()
        
        mode_str = config_dict.get('mode', 'off')
        try:
            mode = SpreadMonitoringMode(mode_str)
        except ValueError:
            # Fall back to OFF if invalid mode string provided
            mode = SpreadMonitoringMode.OFF
            
        return cls(
            mode=mode,
            multiplier=config_dict.get('multiplier', 2.0),
            min_ask_threshold=config_dict.get('min_ask_threshold', 0.30)
        )



"""
This module contains custom exceptions related to trade operations.

OTB-269: Entry retry mechanism - Distinguishes between retryable and fatal trade errors.
"""


class TradeException(Exception):
    """
    Base class for all trade-related exceptions.
    """
    def __init__(self, reason: str, *args):
        """
        Initialize the exception with a reason.

        Args:
            reason: Reason why the trade operation failed
            args: Additional arguments to pass to the base Exception class
        """
        super().__init__(reason, *args)
        self.reason = reason

    def __str__(self):
        return f"Trade operation failed: {self.reason}"


class RetryableTradeException(TradeException):
    """
    Exception raised when a trade operation fails but can be retried.
    
    Examples of retryable failures:
    - Minimum premium not met
    - Maximum premium exceeded
    - Max entry adjustments reached
    - No suitable strikes found (market conditions may change)
    - Adjusted price crosses zero
    """
    def __init__(self, reason: str, *args):
        super().__init__(reason, *args)

    def __str__(self):
        return f"Trade failed (retryable): {self.reason}"


class FatalTradeException(TradeException):
    """
    Exception raised when a trade operation fails and should NOT be retried.
    
    Examples of fatal failures:
    - No broker connection
    - Trading disabled
    - No Hub connection (subscription issue)
    - Maximum open trades reached
    - Template conditions not met (e.g., VIX out of range)
    """
    def __init__(self, reason: str, *args):
        super().__init__(reason, *args)

    def __str__(self):
        return f"Trade failed (fatal): {self.reason}"



import datetime as dt
from abc import ABC, abstractmethod
from bisect import bisect_left
from dataclasses import dataclass
from datetime import date
from typing import Dict, List, final

from eventkit import Event
from loguru import logger

import optrabot.symbolinfo as symbolInfo
from optrabot.broker.optionpricedata import (OptionStrikeData,
                                             OptionStrikePriceData)
from optrabot.broker.order import Execution, OptionRight
from optrabot.broker.order import Order as GenericOrder
from optrabot.broker.order import OrderStatus
from optrabot.config import Config
from optrabot.exceptions import PlaceOrderException
from optrabot.managedtrade import ManagedTrade
from optrabot.models import Account
from optrabot.optionhelper import OptionHelper
from optrabot.tradetemplate.templatefactory import Template


@dataclass
class SymbolData:
	"""
	Holds symbol specific data which are common for all broker connectors
	"""
	def __init__(self):
		self.symbol: str = None
		self.expirations: List[dt.date] = []
		self.lastPrice: float = 0
		self.lastAtmStrike: float = 0
		self.noPriceDataCount = 0
		self.trade_options: bool = True		# True if options trading is enabled for the symbol
		self.dte_expiration_map: Dict[int, dt.date] = {}		# Maps DTE to expiration date
		self.strikes: List[float] = []
		self.optionPriceData: Dict[dt.date, OptionStrikeData] = {}

class BrokerConnector(ABC):
	# Constants for Events
	EVENT_CONNECTED = 'connectedEvent'
	EVENT_CONNECT_FAILED = 'connectFailedEvent'
	EVENT_COMMISSION_REPORT = 'commissionReportEvent'
	EVENT_DISCONNECTED = 'disconnectedEvent'
	EVENT_ORDER_STATUS = 'orderStatusEvent'
	EVENT_ORDER_EXEC_DETAILS = 'orderExecutionDetailsEvent'
	EVENT_BID_ASK_SPREAD_WARNING = 'bidAskSpreadWarningEvent'

	def __init__(self) -> None:
		self._initialized = False
		self._createEvents()
		self.id = None		# ID of the connector
		self.broker = None	# ID of the broker
		self._tradingEnabled = False
		self._managedAccounts: List[Account] = []
		self._last_option_price_update_time: dt.datetime = None
		self._dtes : List[int] = []		# Configured DTEs for option data requests
		self._symbolData: Dict[str, SymbolData] = {}
		self._subscribed_ticker_data = False
		# Multileg order handling capabilities
		self._supports_multileg_orders = True	# True = Broker can process multileg orders natively
		self._internal_order_adjustment = False	# True = Connector handles order adjustments internally
		self._spread_monitoring_config = Config().get_spread_monitoring_config()
		# Cache for active trade strikes (performance optimization)
		self._active_strikes_cache: set[float] = set()
		self._active_strikes_cache_time: dt.datetime | None = None
		pass
	
	@abstractmethod
	async def cancel_order(self, order: GenericOrder):
		"""
		Cancels the given order
		"""
		pass

	@abstractmethod
	async def connect(self):
		""" 
		Establishes a connection to the broker
		"""
		logger.info('Connecting with broker {}', self.id)

	@abstractmethod
	async def disconnect(self):
		""" 
		Disconnects from the broker
		"""
		await self.unsubscribe_ticker_data()
		logger.info('Disconnecting from broker {}', self.id)

	@abstractmethod
	def isConnected(self) -> bool:
		""" 
		Returns True if the broker is connected
		"""
		pass

	@abstractmethod
	def getAccounts(self) -> List[Account]:
		""" 
		Returns the accounts managed by the broker connection
		"""
		pass

	def get_atm_strike(self, symbol:str) -> float:
		""" 
		Returns the ATM strike price for the given symbol based on the buffered option price data.
		If no data is available, it returns None.
		"""
		symbolData = self._symbolData[symbol]
		return symbolData.lastAtmStrike
	
	@abstractmethod
	async def prepareOrder(self, order: GenericOrder, need_valid_price_data: bool = True) -> None:
		"""
		Prepares the given order for execution.
		- Retrieve current market data for order legs

		Raises:
			PrepareOrderException: If the order preparation fails with a specific reason.
		"""

	def oco_as_complex_order(self) -> bool:
		""" 
		Returns True if the broker connection supports OCO orders in form of one complex order
		instead of two single orders
		"""
		return False

	def supports_multileg_orders(self) -> bool:
		"""
		Returns True if the broker can process multileg orders natively (e.g., spreads, iron condors).
		If False, the connector must split multileg orders into individual leg orders.
		"""
		return self._supports_multileg_orders

	def handles_adjustments_internally(self) -> bool:
		"""
		Returns True if the connector handles order price adjustments internally.
		When True, the TradeManager should not call adjustOrder() for this connector.
		This is typically used when the connector splits multileg orders into individual legs
		and needs to adjust each leg independently.
		"""
		return self._internal_order_adjustment

	@abstractmethod
	async def placeOrder(self, managed_trade: ManagedTrade, order: GenericOrder, parent_order: GenericOrder = None) -> None:
		""" 
		Places the given order for a managed account via the broker connection.
		
		Raises:
			PlaceOrderException: If the order placement fails with a specific reason.
		"""
		pass

	@abstractmethod
	async def place_complex_order(self, take_profit_order: GenericOrder, stop_loss_order: GenericOrder, template: Template) -> bool:
		""" 
		Places Take Profit and Stop Loss Order as single complex order
		"""
		pass

	@abstractmethod
	async def adjustOrder(self, managed_trade: ManagedTrade, order: GenericOrder, price: float) -> bool:
		""" 
		Adjusts the given order with the given new price
		"""
		pass

	@abstractmethod
	async def requestTickerData(self, symbols: List[str]) -> None:
		""" 
		Request ticker data for the given symbols and their options
		"""
		pass

	@abstractmethod
	def getFillPrice(self, order: GenericOrder) -> float:
		""" 
		Returns the fill price of the given order if it is filled
		"""
		pass

	def getLastPrice(self, symbol: str) -> float:
		""" 
		Returns the last price of the given symbol
		"""
		pass

	@abstractmethod
	def get_option_strike_data(self, symbol: str, expiration: date) -> OptionStrikeData:
		""" 
		Returns the option strike data for the given symbol and expiration. It is including
		prices and greeks.
		"""
		pass

	@abstractmethod
	def get_option_strike_price_data(self, symbol: str, expiration: date, strike: float) -> OptionStrikePriceData:
		""" 
		Returns the option strike price data for the given symbol, expiration, strike and right
		"""
		pass

	def get_last_option_price_update_time(self) -> dt.datetime:
		"""
		Returns the last update time of the option price data
		"""
		return self._last_option_price_update_time

	@abstractmethod
	def get_strike_by_delta(self, symbol: str, right: str, delta: int) -> float:
		""" 
		Returns the strike price based on the given delta based on the buffered option price data
		"""
		raise NotImplementedError()
	
	@abstractmethod
	def get_strike_by_price(self, symbol: str, right: str, price: float) -> float:
		""" 
		Returns the strike price based on the given premium price based on the buffered option price data
		"""
		raise NotImplementedError()

	def get_min_price_increment(self, price: float) -> float:
		""" 
		Returns the minimum price increment for the given option based on the given price.
		It's based on the information from Tastytrade Help Center article: https://support.tastytrade.com/support/s/solutions/articles/43000435374
		"""
		if price < 3:
			return 0.05
		else:
			return 0.1

	@abstractmethod
	async def eod_settlement_tasks(self) -> None:
		"""
		Perform End of Day settlement tasks
		"""
		pass

	@final
	def set_dtes(self, dtes: List[int]):
		"""
		Sets the configured DTEs for option data requests
		"""
		self._dtes = dtes

	def determine_expiration_date_by_dte(self, symbol: str, dte: int) -> date:
		""" 
		Determines the expiration date for the given symbol based on the given DTE
		"""
		symbolData = self._symbolData[symbol]
		try:
			return symbolData.dte_expiration_map[dte]
		except KeyError:
			return None

	def _determine_relevant_expirations(self, symbolData: SymbolData) -> List[dt.date]:
		"""
		Determines the relevant option expirations for the given symbol data based on the current date and the days to expiration (DTE)
		"""
		current_date = dt.date.today()
		relevant_expirations = []
		for dte in self._dtes:
			# Calculate the expiration date based on the DTE and check if there is an expiration with this DTE, otherwise use the next available expiration
			closest_expiration = None
			for expiration in symbolData.expirations:
				diff = (expiration - current_date).days
				if diff >= dte:
					closest_expiration = expiration
					symbolData.dte_expiration_map[dte] = closest_expiration		# Store mapping of DTE to expiration date
					break
			if closest_expiration and closest_expiration not in relevant_expirations:
				relevant_expirations.append(closest_expiration)
		return relevant_expirations
	
	def _determine_strikes_of_interest(self, symbolData: SymbolData, atmStrike: float) -> List[float]:
		"""
		Determines the strikes of interest based on the given ATM strike price
		"""
		# Die 40 Strikes um den ATM Strike herum abfragen
		number_of_strikes_above = 45
		number_of_strikes_below = 45
		pos = bisect_left(symbolData.strikes, atmStrike)
		lower_bound = max(0, pos - number_of_strikes_below)
		upper_bound = min(len(symbolData.strikes), pos + number_of_strikes_above)
		return symbolData.strikes[lower_bound:upper_bound]
	
	def uses_oco_orders(self) -> bool:
		""" 
		Returns True if the broker connection supports and uses OCO orders for managing take profit and stop loss
		"""
		return False

	def _createEvents(self):
		""" 
		Creates the events for the broker connection
		"""
		self.connectedEvent = Event(self.EVENT_CONNECTED)
		self.disconnectedEvent = Event(self.EVENT_DISCONNECTED)
		self.commissionReportEvent = Event(self.EVENT_COMMISSION_REPORT)
		self.connectFailedEvent = Event(self.EVENT_CONNECT_FAILED)
		self.orderStatusEvent = Event(self.EVENT_ORDER_STATUS)
		self.orderExecutionDetailsEvent = Event(self.EVENT_ORDER_EXEC_DETAILS)
		self.bid_ask_spread_warning_event = Event(self.EVENT_BID_ASK_SPREAD_WARNING)

	def _emit_bid_ask_spread_warning_event(self, symbol: str, strike: float, right: OptionRight, bid_price: float, ask_price: float, spread: float, max_allowed_spread: float) -> None:
		""" 
		Emits the bid-ask spread warning event
		"""
		self.bid_ask_spread_warning_event.emit(symbol, strike, right, bid_price, ask_price, spread, max_allowed_spread)

	def _emitConnectedEvent(self):
		""" 
		Emits the connected event
		"""
		self.connectedEvent.emit(self)

	def _emitDisconnectedEvent(self):
		""" 
		Emits the disconnected event
		"""
		self._managedAccounts = []
		self.disconnectedEvent.emit(self)

	def _emitConnectFailedEvent(self):
		""" 
		Emits the broker connect failed event
		"""
		self.connectFailedEvent.emit(self)

	def _emitCommissionReportEvent(self, order: GenericOrder, execution_id: str, commission: float = 0, fee: float = 0):
		"""
		Emits the commission report event if commission and fee information are delivered
		for a execution which previously has been reported with the according execution_id.
		"""
		self.commissionReportEvent.emit(order, execution_id, commission, fee)

	def _emitOrderExecutionDetailsEvent(self, order: GenericOrder, execution: Execution):
		""" 
		Emits the order execution details event
		"""
		self.orderExecutionDetailsEvent.emit(order, execution)

	def _emitOrderStatusEvent(self, order: GenericOrder, status: OrderStatus, filledAmount: int = 0):
		""" 
		Emits the order status event. Filled amount holds the amount that has been filled with this
		order status change, if the status event is a "Filled" event.
		"""
		self.orderStatusEvent.emit(order, status, filledAmount)

	async def _monitor_bid_ask_spread(
		self, 
		option_strike_price_data: OptionStrikePriceData, 
		generic_symbol: str, 
		strike: float
	) -> None:
		"""
		Monitors the bid-ask spread of the option price data.
		
		Sends a warning notification if the spread exceeds the configured
		threshold (ask/bid ratio > multiplier).
		
		Args:
			option_strike_price_data: Price data for the option strike
			generic_symbol: The generic symbol (e.g., 'SPX')
			strike: The strike price to monitor
			
		Note:
			Behavior is controlled by spread_monitoring config in config.yaml:
			- enabled: Master switch for monitoring
			- mode: 'off', 'active_strikes', or 'all'
			- multiplier: Ask/Bid ratio threshold (default: 2.0)
			- min_ask_threshold: Minimum ask price to trigger check
		"""
		from optrabot.broker.spreadmonitoring import SpreadMonitoringMode

		# Check if monitoring is enabled (mode != OFF)
		if self._spread_monitoring_config.mode == SpreadMonitoringMode.OFF:
			return
		
		# For ACTIVE_STRIKES mode, only monitor strikes in active trades
		if self._spread_monitoring_config.mode == SpreadMonitoringMode.ACTIVE_STRIKES:
			if not self._is_strike_in_active_trade(strike):
				return
		
		# Extract prices with validation
		call_ask_price = self._get_valid_price(option_strike_price_data.callAsk)
		call_bid_price = self._get_valid_price(option_strike_price_data.callBid)
		put_ask_price = self._get_valid_price(option_strike_price_data.putAsk)
		put_bid_price = self._get_valid_price(option_strike_price_data.putBid)

		# Check Call Spread
		self._check_spread_and_emit_warning(
			ask_price=call_ask_price,
			bid_price=call_bid_price,
			symbol=generic_symbol,
			strike=strike,
			right=OptionRight.CALL,
			min_ask_threshold=self._spread_monitoring_config.min_ask_threshold,
			multiplier=self._spread_monitoring_config.multiplier
		)
		
		# Check Put Spread
		self._check_spread_and_emit_warning(
			ask_price=put_ask_price,
			bid_price=put_bid_price,
			symbol=generic_symbol,
			strike=strike,
			right=OptionRight.PUT,
			min_ask_threshold=self._spread_monitoring_config.min_ask_threshold,
			multiplier=self._spread_monitoring_config.multiplier
		)

	def _get_valid_price(self, price: float | None) -> float:
		"""
		Returns a validated price value.
		
		Args:
			price: The price to validate
			
		Returns:
			The price if valid, otherwise 0.0
		"""
		if price is not None and not OptionHelper.isNan(price) and price >= 0:
			return price
		return 0.0

	def _check_spread_and_emit_warning(
		self,
		ask_price: float,
		bid_price: float,
		symbol: str,
		strike: float,
		right: OptionRight,
		min_ask_threshold: float,
		multiplier: float
	) -> None:
		"""
		Checks if the spread exceeds the threshold and emits a warning if so.
		
		Args:
			ask_price: The ask price of the option
			bid_price: The bid price of the option
			symbol: The generic symbol
			strike: The strike price
			right: CALL or PUT
			spread_config: The spread monitoring configuration
		"""
		if ask_price > min_ask_threshold and bid_price > 0:
			spread = round(ask_price - bid_price, 2)
			max_allowed_spread = bid_price * multiplier
			if spread > max_allowed_spread:
				self._emit_bid_ask_spread_warning_event(
					symbol=symbol,
					strike=strike,
					right=right,
					bid_price=bid_price,
					ask_price=ask_price,
					spread=spread,
					max_allowed_spread=max_allowed_spread
				)

	def _is_strike_in_active_trade(self, strike: float) -> bool:
		"""
		Checks if the given strike is part of any active trade's legs.
		
		This is used in ACTIVE_STRIKES monitoring mode to only emit warnings
		for strikes that are relevant to currently open positions.
		
		Uses a cached set of active strikes that is refreshed every 5 seconds
		for performance optimization (avoids iterating all trades on every tick).
		
		Args:
			strike: The strike price to check
			
		Returns:
			True if the strike is part of an active trade, False otherwise
		"""
		CACHE_TTL_SECONDS = 5
		
		# Check if cache needs refresh
		now = dt.datetime.now()
		if (
			self._active_strikes_cache_time is None
			or (now - self._active_strikes_cache_time).total_seconds() > CACHE_TTL_SECONDS
		):
			self._refresh_active_strikes_cache()
			self._active_strikes_cache_time = now
		
		return strike in self._active_strikes_cache
	
	def _refresh_active_strikes_cache(self) -> None:
		"""
		Refreshes the cache of strikes from active trades.
		"""
		from optrabot.trademanager import TradeManager
		try:
			trade_manager = TradeManager()
			new_strikes: set[float] = set()
			for managed_trade in trade_manager.getManagedTrades():
				if managed_trade.isActive():
					for leg in managed_trade.current_legs:
						new_strikes.add(leg.strike)
			self._active_strikes_cache = new_strikes
		except Exception:
			# If TradeManager is not available, clear cache
			self._active_strikes_cache = set()

	def isInitialized(self) -> bool:
		""" 
		Returns True if the broker connector is initialized
		"""
		return self._initialized

	def isTradingEnabled(self) -> bool:
		""" 
		Returns True if trading is enabled
		"""
		return self._tradingEnabled
	
	async def set_trading_enabled(self, enabled: bool, reason: str = None):
		""" 
		Sets the trading enabled flag with an optional reason which is being logged
		"""
		from optrabot.broker.brokerfactory import BrokerFactory
		self._tradingEnabled = enabled
		if reason:
			logger.debug('Trading enabled: {} - Reason: {}', enabled, reason)
			if enabled == False and not BrokerFactory().is_shutting_down() and BrokerFactory().is_market_open():
				from optrabot.tradinghubclient import NotificationType, TradinghubClient
				notification_message = '‼️ Trading with broker ' + self.broker + ' has been disabled'
				if reason:
					notification_message += ' due to: ' + reason 
				notification_message += '!'
				await TradinghubClient().send_notification(NotificationType.WARN, notification_message) 
				logger.error('Trading has been disabled: {}', reason)
		else:
			logger.debug('Trading enabled: {}', enabled)

	@abstractmethod
	async def unsubscribe_ticker_data(self):
		"""
		Unsubscribes from all ticker data from the broker
		"""
		logger.debug(f'Unsubscribing from all ticker data for broker {self.id}')
		self._subscribed_ticker_data = False


"""
Flow Engine Module

This module provides the flow engine functionality for automating
actions based on trading events.
"""

from .flowconfig import FlowActionType
from .flowengine import FlowEngine
from .flowevent import (EarlyExitEventData, FlowEventData, FlowEventType,
                        ManualCloseEventData, StopLossHitEventData,
                        TakeProfitHitEventData, TradeOpenedEventData)

__all__ = [
    'FlowEngine',
    'FlowActionType',
    'FlowEventType',
    'FlowEventData',
    'TradeOpenedEventData',
    'EarlyExitEventData',
    'StopLossHitEventData',
    'TakeProfitHitEventData',
    'ManualCloseEventData'
]
